from ..extension_loader.snapshot import SNAPSHOT_TAKERS
from .command_processor import CommandProcessor

def take_command_processor_snapshot(client):
    """
    Collects all the commands and categories of a client's command processor.
    
    Parameters
    ----------
    client : ``Client``
        The client, who will be snapshotted.
    
    Returns
    -------
    collected : `None` or `tuple` of (`None` or `set` of ``Category``, `None` or `set` of ``Command``)
    """
    command_processor = getattr(client, 'command_processor', None)
    if (command_processor is None) or (not isinstance(command_processor, CommandProcessor)):
        collected = None
    else:
        category_name_to_category = command_processor.category_name_to_category
        if category_name_to_category:
            categories = set(category_name_to_category.values())
        else:
            categories = None
        
        commands = command_processor.commands
        if commands:
            commands = commands.copy()
        else:
            commands = None
        
        if (categories is None) and (commands is None):
            collected = None
        else:
            collected = (commands, categories)
    
    return collected


def calculate_command_processor_snapshot_difference(client, snapshot_old, snapshot_new):
    """
    Calculates the difference between two command processor snapshots.
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    snapshot_old : `None` or `tuple` of (`None` or `set` of ``Category``, `None` or `set` of ``Command``)
        An old snapshot taken.
    snapshot_new : `None` or `tuple` of (`None` or `set` of ``Category``, `None` or `set` of ``Command``)
        A new snapshot.
    
    Returns
    -------
    snapshot_difference : `None` or `tuple` (`tuple` (`None` or `set`, `None` or `set`), \
            `tuple` (`None` or `set`, `None` or `set`))
        The difference between the two snapshots.
    """
    if (snapshot_old is None) or (snapshot_new is None):
        return None
    
    if snapshot_old is None:
        old_categories = None
        old_commands = None
    else:
        old_commands, old_categories = snapshot_old
    
    if snapshot_new is None:
        new_categories = None
        new_commands = None
    else:
        new_commands, new_categories = snapshot_new
    
    
    if (new_commands is None) or (old_commands is None):
        command_interception = None
    else:
        command_interception = old_commands&new_commands
    
    if command_interception is None:
        old_commands = None
        new_commands = None
    else:
        old_commands = old_commands-command_interception
        new_commands = new_commands-command_interception
        
        if not old_commands:
            old_commands = None
        
        if not new_commands:
            new_commands = None
    
    if (old_commands is None) and (new_commands is None):
        command_difference = None
    else:
        command_difference = (old_commands, new_commands)
    
    
    if (old_categories is None) or (new_categories is None):
        category_interception = None
    else:
        category_interception = old_categories&new_categories
    
    if category_interception is None:
        old_categories = None
        new_categories = None
    else:
        old_categories = old_categories-category_interception
        new_categories = new_categories-category_interception
        
        if not old_categories:
            old_categories = None
        
        if not new_categories:
            new_categories = None
    
    if (old_categories is None) and (new_categories is None):
        category_difference = None
    else:
        category_difference = (old_categories, new_categories)
    
    if (command_difference is None) and (category_difference is None):
        snapshot_difference = None
    else:
        snapshot_difference = (command_difference, category_difference)
    
    return snapshot_difference


def revert_command_processor_snapshot(client, snapshot_difference):
    """
    Reverts a snapshot taken from a command processor.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    snapshot_difference : `tuple` (`tuple` (`None` or `set`, `None` or `set`), \
            `tuple` (`None` or `set`, `None` or `set`))
        The taken snapshot.
    """
    command_processor = getattr(client, 'command_processor', None)
    if (command_processor is None) or (not isinstance(command_processor, CommandProcessor)):
        return
    
    command_difference, category_difference = snapshot_difference
    
    if (command_difference is not None):
        old_commands, new_commands = command_difference
        if (new_commands is not None):
            for command in new_commands:
                command_processor._remove_command(command)
        
        if (old_commands is not None):
            for command in old_commands:
                command_processor._add_command(command)
    
    if (category_difference is not None):
        old_categories, new_categories = category_difference
        if (new_categories is not None):
            for category in new_categories:
                command_processor._remove_category(category)
        
        if (old_categories is not None):
            for category in old_categories:
                command_processor._add_category(category)


SNAPSHOT_TAKERS['client.command_processor'] = (
    take_command_processor_snapshot,
    calculate_command_processor_snapshot_difference,
    revert_command_processor_snapshot,
)



__all__ = ()

from functools import partial as partial_func

from ....discord.message import Message
from ....discord.interaction import InteractionEvent, create_auto_custom_id
from ....discord.core import BUILTIN_EMOJIS

EMOJI_LEFT_2 = BUILTIN_EMOJIS['track_previous']
EMOJI_LEFT = BUILTIN_EMOJIS['arrow_backward']
EMOJI_RIGHT = BUILTIN_EMOJIS['arrow_forward']
EMOJI_RIGHT_2 = BUILTIN_EMOJIS['track_next']
EMOJI_CANCEL = BUILTIN_EMOJIS['x']


CUSTOM_ID_CANCEL = create_auto_custom_id()

def default_check(user, event):
    """
    Default check returned by ``get_auto_check`` wrapped into a partial function.
    
    Parameters
    ----------
    user : ``ClientUserBase``
        The user who received the source event.
    event : ``InteractionEvent``
        The received interaction event.
    
    Returns
    -------
    should_process : `bool`
        Whether the menu should process the received interaction.
    """
    if (event.user is user):
        return True
    
    if event.user_permissions.can_manage_messages and (event.interaction.custom_id == CUSTOM_ID_CANCEL):
        return True
    
    return False


def get_auto_check(event):
    """
    If check is not given by the user, tries to auto create one.
    
    Parameters
    ----------
    event : ``InteractionEvent``, ``Message``, `Any`
        Event passed to the constructor.
    
    Returns
    -------
    check : `None` or `functools.partial`
    """
    if isinstance(event, InteractionEvent):
        user = event.user
    elif isinstance(event, Message):
        interaction = event.interaction
        if interaction is None:
            user = None
        else:
            user = interaction.user
    else:
        user = None
    
    if user is None:
        check = None
    else:
        check = partial_func(default_check, event.user)
    
    return check


def top_level_check(self, event):
    """
    Top level check for menus with `.user_check` attribute.
    
    Parameters
    ----------
    self : ``Menu``
        The respective menu instance.
    event : ``InteractionEvent``
        The received interaction event.
    
    Returns
    -------
    should_process : `bool`
        Whether the menu should process the received interaction.
    """
    check = self.user_check
    if check is None:
        should_process = True
    else:
        should_process = check(event)
    
    return should_process


def top_level_get_timeout(self):
    """
    Top level function for menus with `.timeout` attribute.
    
    Attributes
    ----------
    self : `Menu``
        The respective menu instance.
    
    Returns
    -------
    timeout : `float`
    """
    return self.timeout



__all__ = ('MessageReference',)

from ..core import GUILDS, MESSAGES, CHANNELS


class MessageReference:
    """
    A cross guild reference used as a ``Message``'s `.cross_reference` at crosspost messages.
    
    Attributes
    ----------
    _channel : `object`, `None` or ``ChannelBase``
        Internal slot used by the ``.channel`` property.
    _guild : `object`, `None` or ``Guild``
        Internal used by the ``.guild`` property.
    _message : `object`. `None`, ``Message``
        Internal slot used by the ``.message`` property.
    channel_id : `int`
        The referenced message's channel's id. Might be set as `0`.
    guild_id : `int`
        The referenced message's guild's id. Might be set as `None`.
    message_id : `int`
        The referenced message's id. Might be set as `0`.
    """
    __slots__ = ('_channel', '_message', '_guild', 'channel_id', 'guild_id', 'message_id',)
    def __new__(cls, data):
        """
        Creates a ``MessageReference`` from message reference data included inside of a ``Message``'s.
        
        If the message is loaded already, returns that instead.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Message reference data.
        
        Returns
        -------
        self / message : ``MessageReference`` or ``Message``
        """
        message_id = data.get('message_id', None)
        if message_id is None:
            message_id = 0
        else:
            message_id = int(message_id)
            try:
                message = MESSAGES[message_id]
            except KeyError:
                pass
            else:
                return message
        
        channel_id = data.get('channel_id', None)
        if channel_id is None:
            channel_id = None
        else:
            channel_id = int(channel_id)
        
        guild_id = data.get('guild_id', None)
        if guild_id is None:
            guild_id = 0
        else:
            guild_id = int(guild_id)
        
        self = object.__new__(cls)
        
        self.message_id = message_id
        self.channel_id = channel_id
        self.guild_id = guild_id
        self._message = ...
        self._channel = ...
        self._guild = ...
        
        return self
    
    @property
    def channel(self):
        """
        Returns referenced message's channel if found.
        
        Returns
        -------
        channel : `None` or ``ChannelBase`` instance
        """
        channel = self._channel
        if channel is ...:
            channel_id = self.channel_id
            if channel_id:
                channel = CHANNELS.get(channel_id, None)
            else:
                channel = None
            
            self._channel = channel
        
        return channel
    
    @property
    def guild(self):
        """
        Returns referenced message's guild if found.
        
        Returns
        -------
        guild : `None` or ``Guild``
        """
        guild = self.guild
        if guild is ...:
            guild_id = self.guild_id
            if guild_id:
                guild = GUILDS.get(guild_id, None)
            else:
                guild = None
            
            self._guild = guild
        
        return guild
    
    @property
    def message(self):
        """
        Returns referenced message if found.
        
        Returns
        -------
        message : `None` or ``Message``
        """
        message = self.message
        if message is ...:
            message_id = self.message_id
            if message_id:
                message = GUILDS.get(message_id, None)
            else:
                message = None
            
            self._message = message
        
        return message
    
    def __repr__(self):
        """Returns the representation of the message reference."""
        repr_parts = [
            '<',
            self.__class__.__name__,
        ]
        
        message_id = self.message_id
        if message_id:
            repr_parts.append(' message_id=')
            repr_parts.append(repr(message_id))
            field_added = True
        else:
            field_added = False
        
        channel_id = self.channel_id
        if channel_id:
            if field_added:
                repr_parts.append(',')
            else:
                field_added = True
            
            repr_parts.append(' channel_id=')
            repr_parts.append(repr(channel_id))
        
        guild_id = self.guild_id
        if guild_id:
            if field_added:
                repr_parts.append(',')
            
            repr_parts.append(' guild_id=')
            repr_parts.append(repr(guild_id))
        
        repr_parts.append('>')
        
        return ''.join(repr_parts)



__all__ = ('StickerPack', )

from ..core import STICKER_PACKS
from ..bases import DiscordEntity

from .sticker import Sticker


class StickerPack(DiscordEntity, immortal=True):
    """
    A sticker's pack.
    
    Attributes
    ----------
    id : `int`
        The sticker pack's identifier.
    banner_id : `int`
        The banner asset identifier of the sticker pack.
    cover_sticker_id : `int`
        The sticker's identifier, which the pack uses as it's banner. Defaults to `0` if not applicable.
    description : `str`
        The pack's description.
    sku_id : `int`
        The Stock Keeping Unit identifier of the sticker pack.
    stickers : `frozenset` of ``Sticker``
        The stickers of the pack.
    name : `str`
        The name of the sticker pack.
    """
    __slots__ = ('banner_id', 'cover_sticker_id', 'description', 'stickers', 'sku_id', 'name')
    
    def __new__(cls, data):
        """
        Creates a new ``StickerPack`` from the received data.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            sticker-pack data.
        """
        sticker_pack_id = int(data['id'])
        
        try:
            self = STICKER_PACKS[sticker_pack_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sticker_pack_id
            self._set_attributes(data)
        
        return self
    
    
    def __repr__(self):
        """Returns the sticker pack's representation."""
        return f'<{self.__class__.__name__} id={self.id!r}, name={self.name!r}>'
    
    
    def _set_attributes(self, data):
        """
        Sets the sticker pack's attributes.

        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            Sticker-pack data.
        """
        self.name = data['name']
        self.sku_id = int(data['sku_id'])
        
        cover_sticker_id = data.get('cover_sticker_id', None)
        if cover_sticker_id is None:
            cover_sticker_id = 0
        else:
            cover_sticker_id = int(cover_sticker_id)
        self.cover_sticker_id = cover_sticker_id
        
        self.banner_id = int(data['banner_asset_id'])
        self.description = data['description']
        self.stickers = frozenset(Sticker(sticker_data) for sticker_data in data['stickers'])
    
    
    @classmethod
    def _create_and_update(cls, data):
        """
        Updates if exists the sticker pack, created if not.
        
        Parameters
        ----------
        data : `dict` of (`str`, `Any`) items
            sticker-pack data.
            
        Returns
        -------
        self : ``StickerPack``
        """
        sticker_pack_id = int(data['id'])
        
        try:
            self = STICKER_PACKS[sticker_pack_id]
        except KeyError:
            self = object.__new__(cls)
            self.id = sticker_pack_id
        
        self._set_attributes(data)
        
        return self



__all__ = ()

from ...backend.export import include

from ..core import INTERACTION_EVENT_RESPONSE_WAITERS, INTERACTION_EVENT_MESSAGE_WAITERS

Message = include('Message')

def try_resolve_interaction_message(message, interaction):
    """
    Tries to resolve an interaction's message if not yet resolved.
    
    Parameters
    ----------
    message : ``Message``
        Received message.
    interaction : ``MessageInteraction`` or ``InteractionEvent``
        Received message interaction.
    """
    try:
        interaction_event = INTERACTION_EVENT_RESPONSE_WAITERS.pop(interaction.id)
    except KeyError:
        pass
    else:
        interaction_event.message = message
        
        try:
            waiter = INTERACTION_EVENT_MESSAGE_WAITERS[interaction_event]
        except KeyError:
            pass
        else:
            waiter.set_result_if_pending(None)


def process_message_chunk(message_datas, channel):
    """
    Called with the response data after requesting older messages of a channel. It checks whether we can chain
    the messages to the channel's history. If we can it chains them and removes the length limitation too if
    needed.
    
    Parameters
    ----------
    message_datas : `list` of (`dict` of (`str`, `Any`) items) elements
        A list of message's data received from Discord.
    channel : ``ChannelTextBase`` or `None`
        The channels to which the messages are bound to if any.
    
    Returns
    -------
    received : `list` of ``Message`` objects
    """
    received = []
    index = 0
    limit = len(message_datas)
    
    if index != limit:
        if channel is None:
            for message_data in message_datas:
                message = Message(message_data)
                received.append(message)
        else:
            message_data = message_datas[index]
            index += 1
            message, exists = channel._create_find_message(message_data, False)
            received.append(message)
            
            if exists:
                while True:
                    if index == limit:
                        break
                    
                    message_data = message_datas[index]
                    index += 1
                    message, exists = channel._create_find_message(message_data, True)
                    received.append(message)
                    
                    if exists:
                        continue
                    
                    while True:
                        if index == limit:
                            break
                        
                        message_data = message_datas[index]
                        index += 1
                        message = channel._create_old_message(message_data)
                        received.append(message)
                        continue
                    
                    break
            else:
                while True:
                    if index == limit:
                        break
                    
                    message_data = message_datas[index]
                    index += 1
                    message = Message(message_data)
                    received.append(message)
                    continue
    
    return received




from .headers import *
from .http_client import *
from .rate_limit import *
from .rate_limit_groups import *
from .rate_limit_proxy import *
from .urls import *
from . import rate_limit_groups as RATE_LIMIT_GROUPS

__all__ = (
    'RATE_LIMIT_GROUPS',
    *headers.__all__,
    *http_client.__all__,
    *rate_limit.__all__,
    *rate_limit_groups.__all__,
    *rate_limit_proxy.__all__,
    *urls.__all__,
)



__all__ = ()

from ..color import Color


def get_banner_color_from_data(data):
    """
    Gets banner color from the given user data.
    
    Parameters
    ----------
    data : `dict` of (`str`, `Any`) items
        User data.
    
    Returns
    -------
    banner_color : `None` or `str`
    """
    banner_color = data.get('accent_color', None)
    
    if (banner_color is not None):
        banner_color = Color(banner_color)
    
    return banner_color



__all__ = ()

from ...backend.export import include

Client = include('Client')


def filter_clients(clients, flag_mask):
    """
    Filters the clients whether their intents allows the specific flag.
    
    First yields the first client from `clients` what allows the specified flag. If non, then yields `None`.
    If a `None` or not the expected client was yielded, then the generator should be closed.
    
    If the correct client was yielded, then the generator is used at a for loop yielding all the clients from `clients`
    which allow the specified flag including the firstly yielded one.
    
    This function is a generator.
    
    Parameters
    ----------
    clients : `list` of ``Client``
        The clients to filter.
    flag_mask : `int`
        The intent flag's mask based on what the clients will be filtered.
    
    Yields
    -------
    client : ``Client`` or `None`
    """
    index = 0
    limit = len(clients)
    
    while True:
        if index == limit:
            yield None
            return
        
        client = clients[index]
        if client.intents&flag_mask:
            yield client
            break
        
        index += 1
        continue
        
    yield client
    index += 1
    
    while True:
        if index == limit:
            return
        
        client = clients[index]
        if client.intents&flag_mask:
            yield client
        
        index += 1
        continue

def filter_clients_or_me(clients, flag_mask, me):
    """
    Filters the clients whether their intents allow the specific flag. This filter is used, when the clients receive
    the respective event for themselves even if they have the intent disabled.
    
    First yields the first client from `clients` what allows the specified flag.
    
    If non of the clients allow it, then yields `me` (so the source client), what received the event, then expects
    a `user` to be yielded back. At the end when the generator is iterated over inside of a for loop, then it yields
    `me` again (expect if `me` is not the same as the back yielded `user`, but that should not happen, but making sure).
    
    If any of the source clients allow the specified intent flag, then yields the first client what allows it.
    If not the correct client was yielded back, then the generator should be closed. Meanwhile if the correct client
    was yielded, then the generator expects a `user` to be yielded back. After it, the generator is used inside
    of a for loop yielding all the clients from `clients` which allow the specified intent flag including the firstly
    yielded one. At the end yields the received `user` if it is type ``Client`` and it's specified intent flag is not
    allowed.
    
    This function is a generator.
    
    Parameters
    ----------
    clients : `list` of ``Client``
        The clients to filter.
    flag_shift : `int`
        The intent flag's mask based on what the clients will be filtered.
    me : ``Client``
        The source client, what received the respective event.
    
    Yields
    -------
    client : ``Client`` or `None`
    """
    index = 0
    limit = len(clients)
    
    while True:
        if index == limit:
            # If non of the clients have the intent, then yield `me`
            user = yield me
            yield
            if user is me:
                yield me
            return
        
        client = clients[index]
        if client.intents&flag_mask:
            user = yield client
            break
        
        index += 1
        continue
    
    yield
    
    yield client
    index += 1
    
    while True:
        if index == limit:
            break
        
        client = clients[index]
        if client.intents&flag_mask:
            yield client
        
        index += 1
        continue
    
    # Whether the user is type Client and we did not yield it, yield it.
    if not isinstance(user, Client):
        return
    
    if user.intents&flag_mask:
        return
    
    yield user


def filter_just_me(me):
    """
    Yields the source client.
    
    This filter is used when the client itself is the only one who should handle the respective event.
    
    Parameters
    ----------
    me : ``Client``
        The client to yield back.
    
    Yields
    -------
    client : ``Client``
    """
    yield me


def first_client(clients, flag_mask):
    """
    Returns the first client what allows the specified intent flag. If no client allows it, then returns `None`.
    
    Parameters
    ----------
    clients : `list` of ``Client``
        A list of client to search from.
    flag_mask : `int`
        The intent flag's mask based on what the clients will be filtered.
    
    Returns
    -------
    client : ``Client`` or `None`
    """
    index = 0
    limit = len(clients)
    
    while True:
        if index == limit:
            return None
        
        client = clients[index]
        if client.intents&flag_mask:
            return client
            break
        
        index += 1
        continue


def first_client_or_me(clients, flag_mask, me):
    """
    Returns the first client what allows the specified intent flag. If non of the clients allow it, then returns `me`.
    
    Parameters
    ----------
    clients : `list` of ``Client``
        A list of client to search from.
    flag_mask : `int`
        The intent flag's mask based on what the clients will be filtered.
    me : ``Client``
        The source client, what received the respective event.
    
    Returns
    -------
    client : ``Client``
    """
    index = 0
    limit = len(clients)
    
    while True:
        if index == limit:
            return me
        
        client = clients[index]
        if client.intents&flag_mask:
            return client
            break
        
        index += 1
        continue



__all__ = ('Command', )

from ...backend.export import export
from ...backend.utils import WeakReferer
from ...discord.events.handling_helpers import route_value, check_name, Router, route_name, _EventHandlerManager, \
    create_event_from_class
from ...discord.preconverters import preconvert_bool

from .wrappers import CommandWrapper, CommandCheckWrapper
from .category import Category
from .utils import  raw_name_to_display, normalize_description
from .command_helpers import test_error_handler, validate_checks, validate_error_handlers
from .content_parser import CommandContentParser

COMMAND_PARAMETER_NAMES = ('command', 'name', 'description', 'aliases', 'category', 'checks', 'error_handlers',
    'separator', 'assigner', 'hidden', 'hidden_if_checks_fail')

COMMAND_NAME_NAME = 'name'
COMMAND_COMMAND_NAME = 'command'

def _check_maybe_route(variable_name, variable_value, route_to, validator):
    """
    Helper class of ``Command`` parameter routing.
    
    Parameters
    ----------
    variable_name : `str`
        The name of the respective variable
    variable_value : `str`
        The respective value to route maybe.
    route_to : `int`
        The value how much times the routing should happen. by default should be given as `0` if no routing was
        done yet.
    validator : `callable` or `None`
        A callable, what validates the given `variable_value`'s value and converts it as well if applicable.
    
    Returns
    -------
    processed_value : `str`
        Processed value returned by the `validator`. If routing is happening, then a `tuple` of those values is
        returned.
    route_to : `int`
        The amount of values to route to.
    
    Raises
    ------
    ValueError
        Value is routed but to a bad count amount.
    BaseException
        Any exception raised by `validator`.
    """
    if (variable_value is not None) and isinstance(variable_value, tuple):
        route_count = len(variable_value)
        if route_count == 0:
            processed_value = None
        elif route_count == 1:
            variable_value = variable_value[0]
            if variable_value is ...:
                variable_value = None
            
            if validator is None:
                processed_value = variable_value
            else:
                processed_value = validator(variable_value)
        else:
            if route_to == 0:
                route_to = route_count
            elif route_to == route_count:
                pass
            else:
                raise ValueError(f'`{variable_name}` is routed to `{route_count}`, meanwhile something else is '
                    f'already routed to `{route_to}`.')
            
            if validator is None:
                processed_value = variable_value
            else:
                processed_values = []
                for value in variable_value:
                    if (value is not ...):
                        value = validator(value)
                    
                    processed_values.append(value)
                
                processed_value = tuple(processed_values)
    
    else:
        if validator is None:
            processed_value = variable_value
        else:
            processed_value = validator(variable_value)
    
    return processed_value, route_to


def _validate_hidden(hidden):
    """
    Validates the given `is_global` value.
    
    Parameters
    ----------
    hidden : `None` or `bool`
        The `hidden` value to validate.
    
    Returns
    -------
    hidden : `bool`
        The validated `hidden` value.
    
    Raises
    ------
    TypeError
        If `hidden` was not given as `None` nor as `bool` instance.
    """
    if hidden is None:
        hidden = False
    else:
        hidden = preconvert_bool(hidden, 'hidden')
    
    return hidden


def _validate_hidden_if_checks_fail(hidden_if_checks_fail):
    """
    Validates the given `hidden_if_checks_fail` value.
    
    Parameters
    ----------
    hidden_if_checks_fail : `None` or `bool`
        The `hidden_if_checks_fail` value to validate.
    
    Returns
    -------
    hidden_if_checks_fail : `bool`
        The validated `hidden` value.
    
    Raises
    ------
    TypeError
        If `hidden_if_checks_fail` was not given as `None` nor as `bool` instance.
    """
    if hidden_if_checks_fail is None:
        hidden_if_checks_fail = True
    else:
        hidden_if_checks_fail = preconvert_bool(hidden_if_checks_fail, 'hidden_if_checks_fail')
    
    return hidden_if_checks_fail


def _validate_name(name):
    """
    Validates the given name.
    
    Parameters
    ----------
    name : `None` or `str`
        A command's respective name.
    
    Returns
    -------
    name : `None` or `str`
        The validated name.
    
    Raises
    ------
    TypeError
        If `name` is not given as `str` instance.
    """
    if name is not None:
        name_type = name.__class__
        if name_type is str:
            pass
        elif issubclass(name_type, str):
            name = str(name)
        else:
            raise TypeError(f'`name` can be only given as `None` or as `str` instance, got {name_type.__name__}; '
                f'{name!r}.')
    
    return name


def _validate_aliases(aliases):
    """
    Validates the given aliases.
    
    Parameters
    ----------
    aliases : `None`, `str` or `list` of `str`
        Command aliases.
    
    Returns
    -------
    aliases : `None` or `set` of `str`
        The validated aliases.
    
    Raises
    ------
    TypeError
        `aliases` was not given as `None`, `str`, neither as `list` of `str` instances.
    ValueError
        `aliases` contains an empty string.
    """
    if (aliases is not None):
        if isinstance(aliases, list):
            for alias in aliases:
                if not isinstance(alias, str):
                    raise TypeError(f'A non `str` instance alias is given: {alias!r}, got {aliases!r}.')
                
                if not alias:
                    raise ValueError(f'An alias cannot be empty string, got {aliases!r}.')
            
            aliases_processed = set()
            for alias in aliases:
                alias = raw_name_to_display(alias)
                aliases_processed.add(alias)
            
            if aliases_processed:
                aliases = aliases_processed
            else:
                aliases = None
        
        elif isinstance(aliases, str):
            if not aliases:
                raise ValueError(f'An alias cannot be empty string, got {aliases!r}.')
            
            aliases = raw_name_to_display(aliases)
            aliases = {aliases}
        else:
            raise TypeError('Aliases can be given as `str`, or `list` of `str` instances, got '
                f'{aliases.__class__.__name__}; {aliases!r}')
    
    return aliases


def _validate_category(category):
    """
    Validates the given category.
    
    Parameters
    ----------
    category : `None`, `str`  instance or ``Category``
        The category to validate.
    
    Returns
    -------
    category : `str` or ``Category``
        The validated category.
    
    Raises
    ------
    TypeError
        Category is not given either as `None`, `str` instance, or ``Category``.
    """
    if (category is not None):
        category_type = category.__class__
        if category_type is Category:
            pass
        elif category_type is str:
            pass
        elif issubclass(category_type, str):
            category = str(category)
        else:
            raise TypeError(f'`category` should be `None`, type `str` or `{Category.__name__}`, got '
                f'{category_type.__name__}.')
    
    return category


def _generate_description_from(command, description):
    """
    Generates description from the command and it's maybe given description.
    
    Parameters
    ----------
    command : `str`
        The command's function.
    description : `Any`
        The command's description.
    
    Returns
    -------
    description : `str`
        The generated description.
    """
    if description is None:
        description = getattr(command, '__doc__', None)
    
    if (description is not None) and isinstance(description, str):
        description = normalize_description(description)
    
    return description


def _generate_category_hint_from(category):
    """
    Generates category hint from the given category.
    
    Parameters
    ----------
    category : `None`, `str` or ``Category``
        The respective category.
    
    Returns
    -------
    category_hint : `None` or `str`
        The category's string representation if applicable.
    
    Raises
    ------
    TypError
        Category is not given as `None`, `str`, neither as ``Category`` instance.
    """
    if category is None:
        category_hint = None
    else:
        category_type = category.__class__
        if category_type is Category:
            category_hint = category.name
        elif category_type is str:
            category_hint = category
        elif issubclass(category_type, str):
            category_hint = str(category)
        else:
            raise TypeError(f'`category` should be `None`, type `str` or `{Category.__name__}`, got '
                f'{category_type.__name__}.')
    
    return category_hint


def _fetch_check_from_wrappers(wrappers):
    """
    Gets the check wrappers from the given ones.
    
    Parameters
    ----------
    wrappers : `None` or `list` of ``CommandWrapper``
        Command wrappers.
    
    Returns
    -------
    checks : `None` or `list` of ``CheckBase``
        The fetched down checks.
    """
    checks = None
    if (wrappers is not None):
        for wrapper in wrappers:
            if isinstance(wrapper, CommandCheckWrapper):
                if checks is None:
                    checks = []
                
                checks.append(wrapper._check)
    
    return checks


def _iter_merge_checks(checks_1, checks_2):
    """
    Iterates the given checks.
    
    This function is a generator.
    
    Parameters
    ----------
    checks_1 : `None` or `list` of ``CheckBase``
        Checks to merge.
    checks_2: `None` or`list` of ``CheckBase``
        Checks to merge.
    
    Yields
    ------
    check : ``CheckBase``
    """
    if (checks_1 is not None):
        yield from checks_1
    
    if (checks_2 is not None):
        yield from checks_2


def _merge_checks(checks_1, checks_2):
    """
    Merges the two checks.
    
    Parameters
    ----------
    checks_1 : `None` or `list` of ``CheckBase``
        Checks to merge.
    checks_2: `None` or`list` of ``CheckBase``
        Checks to merge.
    
    Returns
    -------
    checks : `None` or `tuple` of ``CheckBase``
        The fetched down checks.
    """
    if (checks_1 is None) and (checks_2 is None):
        return None
    
    return tuple(_iter_merge_checks(checks_1, checks_2))

@export
class Command:
    """
    Represents a command.
    
    Attributes
    ----------
    _category_hint : `None` or `str`
        Hint for the command processor to detect under which category the command should go.
    _category_reference : `None` or ``WeakReferer`` to ``Category``.
        Weak reference to the command's category.
    _checks : `None` or `tuple` of ``CheckBase``
        The checks of the commands.
    _command_categories : `None` or `set` of ``CommandCategory``
        Sub command categories of the command.
    _command_function : `None` or ``CommandFunction``
        The actual command of the command to maybe call.
    _command_processor_reference : `None` or ``WeakReferer`` to ``CommandProcessor``.
        Weak reference to the command's command processor.
    _error_handlers : `None` or `list` of `FunctionType`
        Error handlers bind to the command.
    _self_reference = `None` or ``WeakReferer`` to ``Command``
        Reference to the command itself.
    _wrappers : `None`, `list` of `async-callable`
        Additional wrappers, which run before the command is executed.
    aliases : `None` or `tuple` of `str`
        Name aliases of the command if any. They are always lower case.
    command_category_name_to_command_category : `None` or `dict` of (`str`, ``CommandCategory``) items
        Sub-command categories by name.
    description : `Any`
        The command's description if any.
    display_name : `str`
        The command's display name.
    hidden : `bool`
        Whether the command should be hidden from help commands.
    hidden_if_checks_fail : `bool`
        Whether the command should be hidden from help commands if the user's checks fail.
    name : `str`
        The command's name. Always lower case.
    """
    __slots__ = ('__weakref__', '_category_hint', '_category_reference', '_checks', '_command_categories',
        '_command_function', '_command_processor_reference', '_error_handlers', '_self_reference', '_wrappers',
        'aliases', 'command_category_name_to_command_category', 'description', 'display_name', 'hidden',
        'hidden_if_checks_fail', 'name')
    
    def _iter_checks(self):
        """
        Iterates over all the checks applied to the command.
        
        This method is an iterable generator.
        
        Yields
        ------
        check : ``CheckBase``
        """
        checks = self._checks
        if (checks is not None):
            yield from checks
        
        category_reference = self._category_reference
        if (category_reference is not None):
            category = category_reference()
            if (category is not None):
                checks = category._checks
                if (checks is not None):
                    yield from checks
    
    
    def _iter_own_checks(self):
        """
        Iterates over the checks only of the command, excluding it's category's.
        
        This method is an iterable generator.
        
        Yields
        ------
        check : ``CheckBase``
        """
        checks = self._checks
        if (checks is not None):
            yield from checks
    
    
    def _iter_error_handlers(self):
        """
        Iterates over all the error handlers applied to the command.
        
        This method is an iterable generator.
        
        Yields
        ------
        error_handler : `FunctionType`
        """
        error_handlers = self._error_handlers
        if (error_handlers is not None):
            yield from error_handlers
        
        category_reference = self._category_reference
        if (category_reference is not None):
            category = category_reference()
            if (category is not None):
                error_handlers = category._error_handlers
                if (error_handlers is not None):
                    yield from error_handlers
        
        command_processor_reference = self._command_processor_reference
        if (command_processor_reference is not None):
            command_processor = command_processor_reference()
            if (command_processor is not None):
                error_handlers = command_processor._error_handlers
                if (error_handlers is not None):
                    yield from error_handlers
    
    
    def _iter_names(self):
        """
        Iterates overt he command's names.

        This method is an iterable generator.
        
        Yields
        ------
        command_name : `str`
        """
        yield self.name
        aliases = self.aliases
        if (aliases is not None):
            yield from aliases
    
    
    def _iter_wrappers(self):
        """
        Iterates over the wrappers of the command.
        
        This method is an iterable generator.
        
        Yields
        ------
        wrappers : `Any`
        """
        wrappers = self._wrappers
        if (wrappers is not None):
            yield from wrappers
    
    
    def get_category(self):
        """
        Returns the command's category if has any.
        
        Returns
        -------
        category : `None` or ``Category``
        """
        category_reference = self._category_reference
        if category_reference is None:
            category = None
        else:
            category = category_reference()
        
        return category
    
    
    def unlink_category(self):
        """
        Unlinks the command from it's category and of it's command processor as well.
        """
        command_processor = self.get_command_processor()
        self._command_processor_reference = None
        if (command_processor is not None):
            command_name_to_command = command_processor.command_name_to_command
            for name in self._iter_names():
                try:
                    command = command_name_to_command[name]
                except KeyError:
                    pass
                else:
                    if command is self:
                        del command_name_to_command[name]
            
            command_processor.commands.discard(self)
        
        
        category = self.get_category()
        self._category_reference = None
        if (category is not None):
            category.command_instances.discard(self)
    
    
    def set_category(self, category):
        """
        Sets the command's category.
        
        Parameters
        ----------
        category : ``Category``
            The new category of the command.
        
        Raises
        ------
        RuntimeError
            - The command is bound to a category of an other command processor.
            - The command would only partially overwrite
        """
        self.unlink_category()
        
        category.command_instances.add(self)
        self._category_hint = category.name
        self._category_reference = category._self_reference
        
        command_processor = category.get_command_processor()
        if (command_processor is not None):
            self.set_command_processor(command_processor)
    
    
    def set_command_processor(self, command_processor):
        """
        Sets the command's command processor.
        
        Parameters
        ----------
        command_processor : ``CommandProcessor``
            The command processor to set.
        
        Raises
        ------
        RuntimeError
            - The command is bound to a category of an other command processor.
            - The command would only partially overwrite
        """
        names = set(self._iter_names())
        command_name_to_command = command_processor.command_name_to_command
        
        would_overwrite_commands = set()
        for name in names:
            added_command = command_name_to_command.get(name, None)
            if added_command is None:
                continue
            
            would_overwrite_commands.add(added_command)
        
        for would_overwrite_command in would_overwrite_commands:
            if not names.issubset(would_overwrite_command._iter_names()):
                raise RuntimeError(f'{Command.__name__}: {self!r} would partially overwrite an other command: '
                    f'{would_overwrite_command!r}.')
        
        for name in names:
            command_name_to_command[name] = self
    
    
    def get_command_processor(self):
        """
        Returns the command's command processor if has any.
        
        Returns
        -------
        command_processor : `None` or ``CommandProcessor``
        """
        command_processor_reference = self._command_processor_reference
        if command_processor_reference is None:
            command_processor = None
        else:
            command_processor = command_processor_reference()
        
        return command_processor
    
    
    def error(self, error_handler):
        """
        Adds na error handler to the command.
        
        Parameters
        ----------
        error_handler : `async-callable`
            The error handler to add.
            
            The following parameters are passed to each error handler:
            
            +-------------------+-----------------------+
            | Name              | Type                  |
            +===================+=======================+
            | command_context   | ``CommandContext``    |
            +-------------------+-----------------------+
            | exception         | `BaseException`       |
            +-------------------+-----------------------+
            
            Should return the following parameters:
            
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | handled           | `bool`    |
            +-------------------+-----------+
        
        Returns
        -------
        error_handler : `async-callable`
        
        Raises
        ------
        TypeError
            - If `error_handler` accepts bad amount of parameters.
            - If `error_handler` is not async.
        """
        test_error_handler(error_handler)
        
        error_handlers = self._error_handlers
        if error_handlers is None:
            error_handlers = self._error_handlers = []
            
            error_handlers.append(error_handler)
        
        return error_handler
    
    
    @classmethod
    def from_class(cls, klass, kwargs=None):
        """
        The method used, when creating a ``Command`` object from a class.
        
        Extra `kwargs` are supported as well for special the use cases.
        
        Parameters
        ----------
        klass : `type`
            The class, from what's attributes the command will be created.
        
        Returns
        -------
        self : ``Command`` or ``Router``
        
        Raises
        ------
        TypeError
            If any attribute's or value's type is incorrect.
        ValueError
            If any attribute's or value's type is correct, but it's type isn't.
        """
        return create_event_from_class(cls, klass, COMMAND_PARAMETER_NAMES, COMMAND_NAME_NAME,
            COMMAND_COMMAND_NAME)
    
    
    def __new__(cls, command, name=None, description=None, aliases=None, category=None, checks=None,
            error_handlers=None, separator=None, assigner=None, hidden=None, hidden_if_checks_fail=None):
        """
        Creates a new ``Command`` object.
        
        Parameters
        ----------
        command : `None`, `async-callable`
            The async callable added as the command itself.
        name : `None`, `str` or `tuple` of (`None`, `Ellipsis`, `str`), Optional
            The name to be used instead of the passed `command`'s.
        description : `None`, `Any` or `tuple` of (`None`, `Ellipsis`, `Any`), Optional
            Description added to the command. If no description is provided, then it will check the commands's
            `.__doc__` attribute for it. If the description is a string instance, then it will be normalized with the
            ``normalize_description`` function. If it ends up as an empty string, then `None` will be set as the
            description.
        aliases : `None`, `str`, `list` of `str` or `tuple` of (`None, `Ellipsis`, `str`, `list` of `str`), Optional
            The aliases of the command.
        category : `None`, ``Category``, `str` or `tuple` of (`None`, `Ellipsis`, ``Category``, `str`), Optional
            The category of the command. Can be given as the category itself, or as a category's name. If given as
            `None`, then the command will go under the command processer's default category.
        checks : `None`, ``CommandCheckWrapper``, ``CheckBase``, `list` of ``CommandCheckWrapper``, ``CheckBase`` \
                instances or `tuple` of (`None`, `Ellipsis`, ``CommandCheckWrapper``, ``CheckBase`` or `list` of \
                ``CommandCheckWrapper``, ``CheckBase``), Optional
            Checks to decide in which circumstances the command should be called.
        error_handlers : `None`, `async-callable`, `list` of `async-callable`, `tuple` of (`None`, `async-callable`, \
                `list` of `async-callable`), Optional
            Error handlers for the command.
        separator : `None`, `str` or `tuple` (`str`, `str`), Optional
            The parameter separator of the command's parser.
        assigner : `None`, `str`, Optional
            Parameter assigner sign of the command's parser.
        hidden : `None`, `bool`, `tuple` (`None`, `Ellipsis`, `bool`), Optional
            Whether the command should be hidden from the help commands.
        hidden_if_checks_fail : `None`, `bool`, `tuple` (`None`, `Ellipsis`, `bool`), Optional
            Whether the command should be hidden from the help commands if any check fails.
        """
        if command is None:
            function = None
            wrappers = None
        elif isinstance(command, CommandWrapper):
            function, wrappers = command.fetch_function_and_wrappers_back()
        else:
            function = command
            wrappers = None
        
        fetched_checks = _fetch_check_from_wrappers(wrappers)
        
        route_to = 0
        name, route_to = _check_maybe_route('name', name, route_to, _validate_name)
        description, route_to = _check_maybe_route('description', description, route_to, None)
        aliases, route_to = _check_maybe_route('aliases', aliases, route_to, _validate_aliases)
        category, route_to = _check_maybe_route('category', category, route_to, _validate_category)
        checks, route_to = _check_maybe_route('checks', checks, route_to, validate_checks)
        error_handlers, route_to = _check_maybe_route('error_handlers', error_handlers, route_to,
            validate_error_handlers)
        hidden, route_to = _check_maybe_route('hidden', hidden, route_to, _validate_hidden)
        hidden_if_checks_fail, route_to = _check_maybe_route('hidden_if_checks_fail', hidden_if_checks_fail, route_to,
            _validate_hidden_if_checks_fail)
        
        
        if route_to:
            name = route_name(name, route_to)
            name = [check_name(command, name) for name in name]
            name = [raw_name_to_display(name) for name in name]
            default_description = _generate_description_from(command, None)
            description = route_value(description, route_to, default=default_description)
            aliases = route_value(aliases, route_to)
            category = route_value(category, route_to)
            checks = route_value(checks, route_to)
            error_handlers = route_value(error_handlers, route_to)
            hidden = route_value(hidden, route_to)
            hidden_if_checks_fail = route_value(hidden_if_checks_fail, route_to)
            
            category_hint = [_generate_category_hint_from(category) for category in category]
            
            description = [
                _generate_description_from(function, description)
                    if ((description is None) or (description is not default_description)) else default_description
                for description in description]
        else:
            name = check_name(function, name)
            name = raw_name_to_display(name)
            
            description = _generate_description_from(function, description)
            category_hint = _generate_category_hint_from(category)
        
        if function is None:
            command_function = None
        else:
            content_parser, function = CommandContentParser(function, separator, assigner)
            
            command_function = CommandFunction(function, content_parser)
        
        if route_to:
            router = []
            
            for name, aliases, description, category_hint, checks, error_handlers, hidden, hidden_if_checks_fail \
                    in zip(name, aliases, description, category_hint, checks, error_handlers, hidden,
                        hidden_if_checks_fail):
                
                if (aliases is not None):
                    aliases = tuple(aliases)
                
                self = object.__new__(cls)
                self._category_hint = category_hint
                self._category_reference = None
                self._checks = _merge_checks(checks, fetched_checks)
                self._command_function = command_function
                self._command_processor_reference = None
                self._error_handlers = error_handlers
                self._command_categories = None
                self.aliases = aliases
                self.description = description
                self.display_name = name
                self.hidden = hidden
                self.hidden_if_checks_fail = hidden_if_checks_fail
                self.name = name
                self.command_category_name_to_command_category = None
                self._wrappers = None
                self._self_reference = None
                
                self_reference = WeakReferer(self)
                self._self_reference = self_reference
                
                if (command_function is not None):
                    command_function._command_category_reference = self_reference
                
                
                if (wrappers is not None):
                    for wrapper in wrappers:
                        wrapper.apply(self)
                
                router.append(self)
            
            return Router(router)
        else:
            if (aliases is not None):
                aliases = tuple(aliases)
            
            self = object.__new__(cls)
            self._category_hint = category_hint
            self._category_reference = None
            self._checks = _merge_checks(checks, fetched_checks)
            self._command_function = command_function
            self._command_processor_reference = None
            self._error_handlers = error_handlers
            self._command_categories = None
            self.aliases = aliases
            self.description = description
            self.display_name = name
            self.hidden = hidden
            self.hidden_if_checks_fail = hidden_if_checks_fail
            self.name = name
            self.command_category_name_to_command_category = None
            self._wrappers = None
            self._self_reference = None
            
            self_reference = WeakReferer(self)
            self._self_reference = self_reference
            
            if (command_function is not None):
                command_function._command_category_reference = self_reference
            
            if (wrappers is not None):
                for wrapper in wrappers:
                    wrapper.apply(self)
            
            return self
    
    def __repr__(self):
        """Returns the command's representation."""
        repr_parts = ['<', self.__class__.__name__, ' name=', repr(self.name)]
        
        aliases = self.aliases
        if (aliases is not None):
            repr_parts.append(' aliases=')
            repr_parts.append(repr(aliases))
        
        category_hint = self._category_hint
        if (category_hint is not None):
            repr_parts.append(' category=')
            repr_parts.append(repr(category_hint))
        
        command_function = self._command_function
        if (command_function is not None):
            repr_parts.append(' command_function=')
            repr_parts.append(repr(command_function))
        
        command_categories = self._command_categories
        if (command_categories is not None):
            repr_parts.append(' command_categories=')
            repr_parts.append(repr(command_categories))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    def copy(self):
        """
        Copies the command.
        
        Returns
        -------
        new : ``Command``
        """
        new = object.__new__(type(self))
        new._category_hint = self._category_hint
        new._category_reference = None
        
        checks = self._checks
        if (checks is not None):
            checks = tuple(checks)
        new._checks = checks
        
        command_function = self._command_function
        if (command_function is not None):
            command_function = command_function.copy()
        new._command_function = command_function
        
        new._command_processor_reference = None
        
        error_handlers = self._error_handlers
        if (error_handlers is not None):
            error_handlers = error_handlers.copy()
        new._error_handlers = error_handlers
        
        command_categories  = self._command_categories
        if (command_categories is not None):
            command_categories = {command_category.copy() for command_category in command_categories}
        new._command_categories = command_categories
        
        if (command_categories is None):
            command_category_name_to_command_category = None
        else:
            command_category_name_to_command_category = {}
            for command_category in command_categories:
                for name in command_category._iter_names():
                    command_category_name_to_command_category[name] = command_category
        
        new.command_category_name_to_command_category = command_category_name_to_command_category
        
        aliases = self.aliases
        if (aliases is not None):
            aliases = tuple(aliases)
        new.aliases = aliases
        
        new.description = self.description
        new.display_name = self.display_name
        new.hidden = self.hidden
        new.hidden_if_checks_fail = self.hidden_if_checks_fail
        new.name = self.name
        
        wrappers = self._wrappers
        if (wrappers is not None):
            wrappers = wrappers.copy()
        new._wrappers = wrappers
        
        new._self_reference = None
        
        self_reference = WeakReferer(new)
        self._self_reference = self_reference
        
        if (command_function is not None):
            command_function._command_category_reference = self_reference
        
        if (command_categories is not None):
            for command_category in command_categories:
                command_category._command_category_reference = self_reference
        
        return new
    
    
    def __hash__(self):
        """Returns the hash value of the command."""
        hash_value = 0
        
        category_hint = self._category_hint
        if (category_hint is not None):
            hash_value ^= hash(category_hint)
        
        checks = self._checks
        if (checks is not None):
            hash_value ^= len(checks)<<6
            for check in checks:
                hash_value ^= hash(check)
        
        command_function = self._command_function
        if (command_function is not None):
            try:
                command_function_hash = hash(command_function)
            except TypeError:
                command_function_hash = object.__hash__(command_function)
            hash_value ^= command_function_hash
        
        error_handlers = self._error_handlers
        if (error_handlers is not None):
            hash_value ^= len(error_handlers)<<12
            for error_handler in error_handlers:
                hash_value ^= hash(error_handler)
        
        command_categories = self._command_categories
        if (command_categories is not None):
            hash_value ^= len(command_categories)<<18
            for command_category in command_categories:
                hash_value ^= hash(command_category)
        
        aliases = self.aliases
        if (aliases is not None):
            hash_value ^= len(aliases)<<24
            for alias in aliases:
                hash_value ^= hash(alias)
        
        description = self.description
        if (description is not None):
            try:
                description_hash = hash(description)
            except KeyError:
                description_hash = object.__hash__(description)
            
            hash_value ^= description_hash
        
        hash_value ^= self.hidden<<30
        hash_value ^= self.hidden_if_checks_fail<<31
        hash_value ^= hash(self.name)
        
        wrappers = self._wrappers
        if (wrappers is not None):
            hash_value ^= len(wrappers)<<28
            for wrapper in wrappers:
                try:
                    wrapper_hash = hash(wrapper)
                except TypeError:
                    wrapper_hash = object.__hash__(wrapper)
                hash_value ^= wrapper_hash
        
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two commands are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self._category_hint != other._category_hint:
            return False
        
        if self._checks != other._checks:
            return False
        
        if self._command_function != other._command_function:
            return False
        
        if self._error_handlers != other._error_handlers:
            return False
        
        if self._command_categories != other._command_categories:
            return False
        
        if self.aliases != other.aliases:
            return False
        
        if self.description != other.description:
            return False
        
        if self.hidden != other.hidden:
            return False
        
        if self.hidden_if_checks_fail != other.hidden_if_checks_fail:
            return False
        
        if self.name != other.name:
            return False
        
        if self._wrappers != other._wrappers:
            return False
        
        return True
    
    
    def _add_command_category(self, command_category):
        """
        Adds a command category to the command.
        
        Parameters
        ----------
        command_category : ``CommandCategory``
        """
        command_categories = self._command_categories
        command_category_name_to_command_category = self.command_category_name_to_command_category
        if command_categories is None:
            command_categories = self._command_categories = set()
            command_category_name_to_command_category = self.command_category_name_to_command_category = {}
        else:
            for name in command_category._iter_names():
                if name in command_category_name_to_command_category:
                    raise RuntimeError(f'Duped `{CommandCategory.__name__}` name: {name!r}.')
        
        command_categories.add(command_category)
        for name in command_category._iter_names():
            command_category_name_to_command_category[name] = command_category
        
        command_category._command_category_reference = self._self_reference
    
    
    def create_event(self, command, *args, **kwargs):
        """
        Adds a sub-command to the command.
        
        Parameters
        ----------
        command : ``Command``, ``Router``, `None`, `async-callable`
            Async callable to add as a command.
        *args : Positional parameters
            Positional parameters to pass to the command's constructor.
        **kwargs : Keyword parameters
            Keyword parameters to pass to the command's constructor.
        
        Returns
        -------
        command_category : ``CommandCategory``
            The added command category instance.
        """
        if isinstance(command, (Command, Router)):
            raise TypeError(f'`{Command.__name__}` and `{Router.__name__}` instances cannot be added as sub-commands, '
                f'got {command!r}.')
        
        command = Command(command, *args, **kwargs)
        
        command_category = CommandCategory._from_command(command)
        self._add_command_category(command_category)
        return command_category
    
    
    def create_event_from_class(self, klass):
        """
        Breaks down the given class to it's class attributes and tries to add it as a sub-command.
    
        Parameters
        ----------
        klass : `type`
            The class, from what's attributes the command will be created.
        
        Returns
        -------
        command : ``Command``
            The added command instance.
        """
        command = Command.from_class(klass)
        if isinstance(command, Router):
            command = command[0]
        
        command_category = CommandCategory._from_command(command)
        self._add_command_category(command_category)
        return command_category
    
    
    @property
    def commands(self):
        """
        Enables you to add sub-commands or sub-categories to the command.
        
        Returns
        -------
        handler : ``_EventHandlerManager``
        """
        return _EventHandlerManager(self)
    
    
    def add_wrapper(self, command_wrapper):
        """
        Adds a wrapper to run before the command is executed.
        
        Parameters
        ----------
        command_wrapper : `async-callable`
            Command wrapper to add.
        """
        wrappers = self._wrappers
        if (wrappers is None):
            wrappers = self._wrappers = []
        
        wrappers.append(command_wrapper)
    
    def add_check(self, check):
        """
        Adds a check to the command.
        
        Parameters
        ----------
        check : ``CheckBase``
        """
        checks = self._checks
        if checks is None:
            checks = (check,)
        else:
            checks = (*checks, check)
        
        self._checks = checks

class CommandFunction:
    """
    Command function.
    
    Attributes
    ----------
    _command_category_reference : `None` or ``WeakReferer`` to (``Command`` or ``CommandCategory``)
        Reference to the command function's parent.
    _function : `Any`
        The command's function to call.
    _content_parser : ``CommandContentParser``
        Content parser for the command.
    """
    __slots__ = ('_command_category_reference', '_content_parser', '_function', )
    def __new__(cls, function, content_parser):
        """
        Creates a new ``CommandFunction`` with the given parameters.
        
        Parameters
        ----------
        function : `Any`
            Content parser for the command.
        content_parser : ``CommandContentParser``
            The command's function to call.
        """
        self = object.__new__(cls)
        self._function = function
        self._content_parser = content_parser
        self._command_category_reference = None
        return self
    
    def __repr__(self):
        """Returns the command function's representation."""
        return f'<{self.__class__.__name__} content_parser={self._content_parser!r} function={self._function!r}>'
    
    def copy(self):
        """
        Copying a command function returns itself.
        
        Returns
        -------
        new : ``CommandFunction``
        """
        return self
    
    def __hash__(self):
        """Returns the hash value of the command function."""
        return hash(self._content_parser)^hash(self._function)
    
    def __eq__(self, other):
        """Returns whether the wo command functions are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self is other:
            return True
        
        if self._function != other._function:
            return False
        
        if self._content_parser != other._content_parser:
            return False
        
        return True

    def _iter_wrappers(self):
        """
        Iterates over the wrappers of the command function and of it's parent categories.
        
        This method is an iterable generator.
        
        Yields
        ------
        wrapper : `Any`
        """
        command_category_reference = self._command_category_reference
        if (command_category_reference is not None):
            command_category = command_category_reference()
            if (command_category is not None):
                yield from command_category._iter_wrappers()
    
    
    def _iter_error_handlers(self):
        """
        Iterates over all the error handlers applied to the command function's parent categories..
        
        This method is an iterable generator.
        
        Yields
        ------
        error_handler : `FunctionType`
        """
        command_category_reference = self._command_category_reference
        if (command_category_reference is not None):
            command_category = command_category_reference()
            if (command_category is not None):
                yield from command_category._iter_error_handlers()


class CommandCategory:
    """
    Folder for sub-commands of a ``Command``.
    
    Attributes
    ----------
    _command_categories : `None` or `set` of ``CommandCategory``
        Sub command categories of the command.
    _command_category_reference : `None` or ``WeakReferer`` to (``Command`` or ``CommandCategory``)
        Reference to the command category's parent.
    _command_function : `None` or ``CommandFunction``
        The actual command of the command to maybe call.
    _self_reference : `None` or ``WeakReferer`` to ``CommandCategory``
        Reference to the command category itself.
    _error_handlers : `None` or `list` of `FunctionType`
        Error handlers bind to the command.
    _wrappers : `None`, `list` of `async-callable`
        Additional wrappers, which run before the command is executed.
    aliases : `None` or `tuple` of `str`
        Name aliases of the command category.
    command_category_name_to_command_category : `None` of `dict` of (`str`, `Any`) items
        Command categories by name.
    description : `Any`
        The command's description if any.
    display_name : `str`
        The command's display name.
    name : `str`
        The command's name. Always lower case.
    """
    __slots__ = ('__weakref__', '_command_categories', '_command_category_reference', '_command_function',
        '_error_handlers', '_self_reference', '_wrappers', 'aliases', 'command_category_name_to_command_category',
        'description', 'display_name', 'name',)
    
    @classmethod
    def _from_command(cls, source_command):
        """
        Creates command category from the given command.
        
        Parameters
        ----------
        source_command : ``Command``
            The command to turn into command category.
        
        Returns
        -------
        self : ``CommandCategory``
        """
        self = object.__new__(cls)
        self.aliases = source_command.aliases
        self.name = source_command.name
        self.display_name = source_command.display_name
        self.description = source_command.description
        command_function = source_command._command_function
        self._command_function = command_function
        self._command_categories = source_command._command_categories
        self.command_category_name_to_command_category = None
        self._command_category_reference = None
        self._self_reference = None
        
        wrappers = source_command._wrappers
        if (wrappers is not None):
            wrappers = wrappers.copy()
        self._wrappers = wrappers
        
        error_handlers = source_command._error_handlers
        if (error_handlers is not None):
            error_handlers = error_handlers.copy()
        self._error_handlers = error_handlers
        
        self_reference = WeakReferer(self)
        self._self_reference = self_reference
        
        if (command_function is not None):
            command_function._command_category_reference = self_reference
        
        return self
    
    def copy(self):
        """
        Copies the command category.
        
        Returns
        -------
        new : ``CommandCategory``
        """
        new = object.__new__(type(self))
        
        command_function = self._command_function
        if (command_function is not None):
            command_function = command_function.copy()
        new._command_function = command_function
        
        error_handlers = self._error_handlers
        if (error_handlers is not None):
            error_handlers = error_handlers.copy()
        new._error_handlers = error_handlers
        
        command_categories  = self._command_categories
        if (command_categories is not None):
            command_categories = {command_category.copy() for command_category in command_categories}
        new._command_categories = command_categories
        
        if (command_categories is None):
            command_category_name_to_command_category = None
        else:
            command_category_name_to_command_category = {}
            for command_category in command_categories:
                for name in command_category._iter_names():
                    command_category_name_to_command_category[name] = command_category
        
        new.command_category_name_to_command_category = command_category_name_to_command_category
        
        aliases = self.aliases
        if (aliases is not None):
            aliases = tuple(aliases)
        new.aliases = aliases
        
        new.description = self.description
        new.display_name = self.display_name
        new.name = self.name
        
        wrappers = self._wrappers
        if (wrappers is not None):
            wrappers = wrappers.copy()
        new._wrappers = wrappers
        
        new._command_category_reference = self._command_category_reference
        
        new._self_reference = None
        
        self_reference = WeakReferer(new)
        self._self_reference = self_reference
        
        if (command_function is not None):
            command_function._command_category_reference = self_reference
        
        if (command_categories is not None):
            for command_category in command_categories:
                command_category._command_category_reference = self_reference
        
        return new
    
    
    def __repr__(self):
        """Returns the command category's representation."""
        repr_parts = ['<', self.__class__.__name__, ' name=', repr(self.name)]
        
        command_function = self._command_function
        if (command_function is not None):
            repr_parts.append(' command_function=')
            repr_parts.append(repr(command_function))
        
        command_categories = self._command_categories
        if (command_categories is not None):
            repr_parts.append(' command_categories=')
            repr_parts.append(repr(command_categories))
        
        repr_parts.append('>')
        return ''.join(repr_parts)
    
    def __hash__(self):
        """Returns the command category's hash value."""
        hash_value = 0
        
        command_function = self._command_function
        if (command_function is not None):
            try:
                command_function_hash = hash(command_function)
            except TypeError:
                command_function_hash = object.__hash__(command_function)
            hash_value ^= command_function_hash
        
        error_handlers = self._error_handlers
        if (error_handlers is not None):
            hash_value ^= len(error_handlers)<<12
            for error_handler in error_handlers:
                hash_value ^= hash(error_handler)
        
        command_categories = self._command_categories
        if (command_categories is not None):
            hash_value ^= len(command_categories)<<18
            for command_category in command_categories:
                hash_value ^= hash(command_category)
        
        aliases = self.aliases
        if (aliases is not None):
            hash_value ^= len(aliases)<<24
            for alias in aliases:
                hash_value ^= hash(alias)
        
        description = self.description
        if (description is not None):
            try:
                description_hash = hash(description)
            except KeyError:
                description_hash = object.__hash__(description)
            
            hash_value ^= description_hash
        
        wrappers = self._wrappers
        if (wrappers is not None):
            hash_value ^= len(wrappers)<<28
            for wrapper in wrappers:
                try:
                    wrapper_hash = hash(wrapper)
                except TypeError:
                    wrapper_hash = object.__hash__(wrapper)
                hash_value ^= wrapper_hash
        
        hash_value ^= hash(self.name)
        return hash_value
    
    
    def __eq__(self, other):
        """Returns whether the two command categories are equal."""
        if type(self) is not type(other):
            return NotImplemented
        
        if self._command_function != other._command_function:
            return False
        
        if self._error_handlers != other._error_handlers:
            return False
        
        if self._command_categories != other._command_categories:
            return False
        
        if self.description != other.description:
            return False
        
        if self.name != other.name:
            return False
        
        if self._wrappers != other._wrappers:
            return False
        
        return True
    
    
    def _iter_names(self):
        """
        Iterates over the command category's names.
        
        This method is a generator.
        
        Yields
        ------
        name : `str`
        """
        yield self.name
        
        aliases = self.aliases
        if (aliases is not None):
            yield from aliases
    
    
    def _iter_wrappers(self):
        """
        Iterates over the wrappers of the command category and of parent categories.
        
        This method is an iterable generator.
        
        Yields
        ------
        wrapper : `Any`
        """
        wrappers = self._wrappers
        if (wrappers is not None):
            yield from wrappers
        
        command_category_reference = self._command_category_reference
        if (command_category_reference is not None):
            command_category = command_category_reference()
            if (command_category is not None):
                yield from command_category._iter_wrappers()
    
    
    def _iter_error_handlers(self):
        """
        Iterates over all the error handlers applied to the command categories's parent categories..
        
        This method is an iterable generator.
        
        Yields
        ------
        error_handler : `FunctionType`
        """
        error_handlers = self._error_handlers
        if (error_handlers is not None):
            yield from error_handlers
        
        command_category_reference = self._command_category_reference
        if (command_category_reference is not None):
            command_category = command_category_reference()
            if (command_category is not None):
                yield from command_category._iter_error_handlers()
    
    
    def _add_command_category(self, command_category):
        """
        Adds a command category to the command.
        
        Parameters
        ----------
        command_category : ``CommandCategory``
        """
        command_categories = self._command_categories
        command_category_name_to_command_category = self.command_category_name_to_command_category
        if command_categories is None:
            command_categories = self._command_categories = set()
            command_category_name_to_command_category = self.command_category_name_to_command_category = {}
        else:
            for name in command_category._iter_names():
                if name in command_category_name_to_command_category:
                    raise RuntimeError(f'Duped `{CommandCategory.__name__}` name: {name!r}.')
        
        command_categories.add(command_category)
        for name in command_category._iter_names():
            command_category_name_to_command_category[name] = command_category
        
        command_category._command_category_reference = self._self_reference
    
    
    def create_event(self, command, *args, **kwargs):
        """
        Adds a sub-command to the command.
        
        Parameters
        ----------
        command : ``Command``, ``Router``, `None`, `async-callable`
            Async callable to add as a command.
        *args : Positional parameters
            Positional parameters to pass to the command's constructor.
        **kwargs : Keyword parameters
            Keyword parameters to pass to the command's constructor.
        
        Returns
        -------
        command_category : ``CommandCategory``
            The added command category instance.
        """
        if isinstance(command, (Command, Router)):
            raise TypeError(f'`{Command.__name__}` and `{Router.__name__}` instances cannot be added as sub-commands, '
                f'got {command!r}.')
        
        command = Command(command, *args, **kwargs)
        
        command_category = CommandCategory._from_command(command)
        self._add_command_category(command_category)
    
    
    def create_event_from_class(self, klass):
        """
        Breaks down the given class to it's class attributes and tries to add it as a sub-command.
    
        Parameters
        ----------
        klass : `type`
            The class, from what's attributes the command will be created.
        
        Returns
        -------
        command : ``Command``
            The added command instance.
        """
        command = Command.from_class(klass)
        if isinstance(command, Router):
            command = command[0]
        
        command_category = CommandCategory._from_command(command)
        self._add_command_category(command_category)
        return command_category
    
    
    @property
    def commands(self):
        """
        Enables you to add sub-commands or sub-categories to the command.
        
        Returns
        -------
        handler : ``_EventHandlerManager``
        """
        return _EventHandlerManager(self)
    
    
    def error(self, error_handler):
        """
        Adds na error handler to the command.
        
        Parameters
        ----------
        error_handler : `async-callable`
            The error handler to add.
            
            The following parameters are passed to each error handler:
            
            +-------------------+-----------------------+
            | Name              | Type                  |
            +===================+=======================+
            | command_context   | ``CommandContext``    |
            +-------------------+-----------------------+
            | exception         | `BaseException`       |
            +-------------------+-----------------------+
            
            Should return the following parameters:
            
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | handled           | `bool`    |
            +-------------------+-----------+
        
        Returns
        -------
        error_handler : `async-callable`
        
        Raises
        ------
        TypeError
            - If `error_handler` accepts bad amount of parameters.
            - If `error_handler` is not async.
        """
        test_error_handler(error_handler)
        
        error_handlers = self._error_handlers
        if error_handlers is None:
            error_handlers = self._error_handlers = []
            
            error_handlers.append(error_handler)
        
        return error_handler



__all__ = ('Preinstance', 'PreinstancedBase', )

class Preinstance:
    """
    name : `str`
        The instance's name.
    value : `str` or `int`
        The instance's value.
    args : `tuple` of `Any`
        Additional parameters to preinstance with.
    """
    __slots__ = ('name', 'value', 'args')
    
    def __new__(cls, value, name, *args):
        """
        Creates a new ``Preinstance`` instance with the given parameters.
        
        Parameters
        ----------
        value : `str` or `int`
            The instance's value.
        name : `str`
            The instance's name.
        *args : Parameters
            Additional parameters to preinstance with.
        """
        self = object.__new__(cls)
        self.name = name
        self.value = value
        self.args = args
        return self
    
    def __repr__(self):
        """Returns the preinstanced's representation."""
        repr_parts = [
            self.__class__.__name__, '(',
            repr(self.value), ', ', repr(self.name),
        ]
        
        args = self.args
        for arg in args:
            repr_parts.append(', ')
            repr_parts.append(repr(arg))
        
        repr_parts.append(')')
        
        return ''.join(repr_parts)


class PreinstancedMeta(type):
    """
    Metaclass for ``PreinstancedBase`` instances.
    """
    def __new__(cls, class_name, class_parents, class_attributes):
        """
        Creates a preinstanced type.
        
        Parameters
        ----------
        class_name : `str`
            The created class's name.
        class_parents : `tuple` of `type` instances
            The superclasses of the creates type.
        class_attributes : `dict` of (`str`, `Any`) items
            The class attributes of the created type.
        
        Returns
        -------
        type : ``PreinstancedMeta`` instance
        """
        post_instance = []
        new_class_attributes = {}
        for attribute_name, attribute_value in class_attributes.items():
            if isinstance(attribute_value, Preinstance):
                post_instance.append((attribute_name, attribute_value))
            else:
                new_class_attributes[attribute_name] = attribute_value
        
        type_ = type.__new__(cls, class_name, class_parents, new_class_attributes)
        
        for attribute_name, attribute_value in post_instance:
            setattr(type_, attribute_name, type_(attribute_value.value, attribute_value.name, *attribute_value.args)),
        
        return type_


class PreinstancedBase(metaclass=PreinstancedMeta):
    """
    Base class for preinstanced types.
    
    Class Attributes
    ----------------
    INSTANCES : `NoneType` = `NotImplemented`
        The instances of the preinstanced type. Subclasses should overwrite it as `dict`.
    VALUE_TYPE : `type` = `NoneType`
        The preinstanced object's value's type. Subclasses should overwrite it.
    DEFAULT_NAME : `str` = `'Undefined'`
        The default name to use as the preinstanced objects'.
    """
    INSTANCES = NotImplemented
    VALUE_TYPE = None.__class__
    DEFAULT_NAME = 'Undefined'
    
    __slots__ = ('name', 'value',)
    
    @classmethod
    def get(cls, value):
        """
        Returns the value's representation. If the value is already preinstanced, returns that, else creates a new
        object.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value to get it's representation.
        
        Returns
        -------
        obj_ : ``PreinstancedBase`` instance
        """
        try:
            obj_ = cls.INSTANCES[value]
        except KeyError:
            obj_ = cls._from_value(value)
        
        return obj_
    
    
    @classmethod
    def _from_value(cls, value):
        """
        Creates a new preinstanced object from the given value.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value what has no representation yet.
        
        Returns
        -------
        self : ``PreinstancedBase`` instance
            The created object.
        """
        self = object.__new__(cls)
        self.value = value
        self.name = cls.DEFAULT_NAME
        self.INSTANCES[value] = self
        return self
    
    
    def __init__(self, value, name):
        """
        Creates a new preinstanced instance.
        
        Parameters
        ----------
        value : ``.VALUE_TYPE``
            The value of the preinstanced object.
        name : `str`
            The object's name.
        """
        self.value = value
        self.name = name
        self.INSTANCES[value] = self
    
    
    def __gt__(self, other):
        """Returns whether self's value is greater than the other object's."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        if self.value > other_value:
            return True
        else:
            return False
    
    
    def __ge__(self, other):
        """Returns whether self's value is greater or equal to the other object's."""
        if self is other:
            return True
        
        other_type = other.__class__
        self_type = self.__class__
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        if self.value >= other_value:
            return True
        else:
            return False
    
    
    def __eq__(self, other):
        """Returns whether self's value equals to the other object's."""
        if self is other:
            return True
        
        other_type = other.__class__
        self_type = self.__class__
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        if self.value == other_value:
            return True
        else:
            return False
    
    
    def __ne__(self, other):
        """Returns whether self's not equals to the other object's."""
        if self is other:
            return False
        
        other_type = other.__class__
        self_type = self.__class__
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        if self.value != other_value:
            return True
        else:
            return False
    
    
    def __le__(self, other):
        """Returns whether self's value is less or equal to the other object's."""
        if self is other:
            return True
        
        other_type = other.__class__
        self_type = self.__class__
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        if self.value <= other_value:
            return True
        else:
            return False
    
    
    def __lt__(self, other):
        """Returns whether self's value is less than the other object's."""
        other_type = other.__class__
        self_type = self.__class__
        if other_type is self_type:
            other_value = other.value
        elif other_type is self_type.VALUE_TYPE:
            other_value = other
        else:
            return NotImplemented
        
        if self.value < other_value:
            return True
        else:
            return False
    
    
    def __hash__(self):
        """Returns the hash of the preinstanced object."""
        return hash(self.value)
    
    
    def __repr__(self):
        """Returns the representation of the preinstanced object."""
        return f'{self.__class__.__name__}(value={self.value!r}, name={self.name!r})'



__all__ = ()

from .rate_limit import RateLimitGroup, LIMITER_WEBHOOK, LIMITER_CHANNEL, StaticRateLimitGroup, LIMITER_GUILD, \
    LIMITER_INTERACTION

__doc__ = """
Defines the rate limit groups by hata. Hata uses burst half automatic rate limit handler.

Burst rate limit handler means, if (for example) 30 requests can be done to an endpoint before it resets, it will
let 30 requests to pass trough, and with 31th will wait till the limits expires.

Half automatic, since it automatically detects rate limit sizes with it's first request, but it do not detects
which endpoints are limited together and by which id, since they are set by their rate limit group.

It is optimistic, since Discord do not limits every endpoint, but endpoints which will be potentially changed
are marked as optimistic. They have a limiter set, but their limit can be subject of change. The implementation
lets trough 1 more request with each cycle, starting from `1`, till it reaches a set `n` limit. If any of the
requests return rate limit information, then changes the endpoints limitations to it. The increased pararellity
starting by `1` is to ensure, that the endpoint is mapped by first request. The increasing pararellity represents
the decreasing chance of getting any rate limit information back.

Limit Guides
------------
The following shortenings are used inside of group descriptions:
- `N/A` : Not applicable
- `UN` : Unknown
- `OPT` : Optimistic

Limiter types:
- `UNLIMITED`
- `GLOBAL`
- `channel_id`
- `guild_id`
- `interaction_id`
- `webhook_id`

Required auth types:
- `N/A` (No auth required.)
- `UN` (Unknown, not bot.)
- `application`
- `bearer`
- `bot`
- `user`

Content Delivery Network Endpoint
---------------------------------
- Endpoint : `https://cdn.discordapp.com/`
- Method : `GET`
- Required auth : `N/A`
- Limiter : `UNLIMITED`
- Limit : `N/A`
- Resets after : `N/A`


Status Endpoints
----------------
- Endpoint base: `https://status.discord.com/api/v2`
- endpoints:
    - `/incidents/unresolved.json`
    - `/scheduled-maintenances/active.json`
    - `/scheduled-maintenances/upcoming.json`
- Method : `GET`
- Required auth : `N/A`
- Limiter : `UNLIMITED`
- Limit : `N/A`
- Resets after : `N/A`

Shared Groups
-------------
- GROUP_REACTION_MODIFY
    - Used by : `reaction_clear`, `reaction_delete_emoji`, `reaction_delete_own`, `reaction_add`, `reaction_delete`
    - Limiter : `channel_id`
    - Limit : `1`
    - Resets after : `0.25`

- GROUP_PIN_MODIFY
    - Used by : `message_unpin`, `message_pin`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `4.0`

- GROUP_USER_MODIFY
    - Used by : `user_edit`, `user_move`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- GROUP_USER_ROLE_MODIFY
    - Used by : `user_role_delete`, `user_role_add`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- GROUP_WEBHOOK_EXECUTE
    - Used by : `webhook_message_create`, `webhook_message_delete`, `webhook_message_get`, `webhook_message_edit`
    - Limiter : `webhook_id`
    - Limit : `5`
    - Resets after : `2.0`

- GROUP_INTERACTION_EXECUTE
    - Used by: `interaction_followup_message_create`, `interaction_response_message_delete`,
        `interaction_response_message_edit`, `interaction_followup_message_delete`,
        `interaction_response_message_get`, `interaction_followup_message_edit`, `interaction_followup_message_get`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- GROUP_PERMISSION_OVERWRITE_MODIFY
    - Used by: `permission_overwrite_delete`, `permission_overwrite_create`
    - Limiter : `channel_id`
    - Limit : `10`
    - Resets after : `15.0`

- GROUP_THREAD_CREATE
    - Used by: `thread_create_from_message`, `thread_create`
    - Limiter : `GLOBAL`
    - Limit : `50`
    - Resets after : `300.0`

- GROUP_THREAD_ACTION
    - Used by: `thread_join`, `thread_leave`, `thread_user_add`, `thread_user_delete`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

Group Details
-----------
- oauth2_token
    - Endpoint : `oauth2/token`
    - Method : `POST`
    - Required auth : `application`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- application_get
    - Endpoint : `/applications/{application_id}`
    - Method : `GET`
    - Required auth : `UN`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- achievement_get_all
    - Endpoint : `/applications/{application_id}/achievements`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- achievement_create
    - Endpoint : `/applications/{application_id}/achievements`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- achievement_delete
    - Endpoint : `/applications/{application_id}/achievements/{achievement_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- achievement_get
    - Endpoint : `/applications/{application_id}/achievements/{achievement_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- achievement_edit
    - Endpoint : `/applications/{application_id}/achievements/{achievement_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- application_command_global_get_all
    - Endpoint : `/applications/{application_id}/commands`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- application_command_global_create
    - Endpoint : `/applications/{application_id}/commands`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `20.0`

- application_command_global_update_multiple
    - Endpoint : `/applications/{application_id}/commands`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `2`
    - Resets after : `60.0`

- application_command_global_get
    - Endpoint : `/applications/{application_id}/commands/{application_command_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- application_command_global_delete
    - Endpoint : `/applications/{application_id}/commands/{application_command_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `20.0`

- application_command_global_edit
    - Endpoint : `/applications/{application_id}/commands/{application_command_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `20.0`

- application_command_guild_get_all
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- application_command_guild_create
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `5`
    - Resets after : `20.0`

- application_command_guild_update_multiple
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `2`
    - Resets after : `60.0`

- application_command_permission_get_all_guild
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands/permissions'`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- application_command_guild_get
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands/{application_command_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- application_command_guild_delete
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands/{application_command_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `5`
    - Resets after : `20.0`

- application_command_guild_edit
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands/{application_command_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `5`
    - Resets after : `20.0`

- application_command_permission_get
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands/{application_command_id}/permissions`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- application_command_permission_edit
    - Endpoint : `/applications/{application_id}/guilds/{guild_id}/commands/{application_command_id}/permissions`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5.0`
    - Resets after : `20.0`

- application_get_all_detectable
    - Endpoint : `/applications/detectable`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- client_logout
    - Endpoint : `/auth/logout`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `N/A`
    - Resets after : `N/A`
    - Notes : Untested.

- channel_delete
    - Endpoint : `/channels/{channel_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- channel_group_leave
    - Endpoint : `/channels/{channel_id}`
    - Method : `DELETE`
    - Required auth : `user`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`
    - Notes : Untested.

- channel_edit
    - Endpoint : `/channels/{channel_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `15`
    - Notes : Has sub-limits.

- channel_group_edit
    - Endpoint : `/channels/{channel_id}`
    - Method : `PATCH`
    - Required auth : `user`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- channel_directory_counts
    - Endpoint : `/channels/{channel_id}/directory-entries/counts`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- channel_directory_get_all
    - Endpoint : `/channels/{channel_id}/directory-entries/list`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- channel_directory_search
    - Endpoint : `/channels/{channel_id}/directory-entries/search`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- channel_follow
    - Endpoint : `/channels/{channel_id}/followers`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- greet
    - Endpoint : `/channels/{channel_id}/greet`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- invite_get_all_channel
    - Endpoint : `/channels/{channel_id}/invites`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- invite_create
    - Endpoint : `/channels/{channel_id}/invites`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `15.0`

- message_get_chunk
    - Endpoint : `/channels/{channel_id}/messages`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `5.0`

- message_create
    - Endpoint : `/channels/{channel_id}/messages`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `4.0`

- message_delete_multiple
    - Endpoint : `/channels/{channel_id}/messages/bulk-delete`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `1`
    - Resets after : `3.0`

- message_delete
    - Endpoint : `/channels/{channel_id}/messages/{message_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `3`
    - Resets after : `1.0`
    - Notes : Applicable for messages posted by the bot or which are younger than 2 weeks. Has sub-limits.

- message_delete_b2wo
    - Endpoint : `/channels/{channel_id}/messages/{message_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `30`
    - Resets after : `120.0`
    - Notes : Applicable for messages which are not posted by the bot and are older than 2 weeks. Has sub-limits.

- message_get
    - Endpoint : `/channels/{channel_id}/messages/{message_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `5.0`

- message_edit
    - Endpoint : `/channels/{channel_id}/messages/{message_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `4.0`

- message_ack
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/ack`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- message_crosspost
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/crosspost`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `10`
    - Resets after : `3600.0`

- message_interaction
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/interaction-data`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`

- reaction_clear
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/reactions`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `1`
    - Resets after : `0.25`

- reaction_delete_emoji
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/reactions/{reaction}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `1`
    - Resets after : `0.25`

- reaction_user_get_chunk
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/reactions/{reaction}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- reaction_delete_own
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/reactions/{reaction}/@me`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `1`
    - Resets after : `0.25`

- reaction_add
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/reactions/{reaction}/@me`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `1`
    - Resets after : `0.25`

- reaction_delete
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/reactions/{reaction}/{user_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `1`
    - Resets after : `0.25`

- message_suppress_embeds
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/suppress-embeds`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `3`
    - Resets after : `1`

- thread_create_from_message
    - Endpoint : `/channels/{channel_id}/messages/{message_id}/threads`
    - Method : `POST`
    - Required auth : `UN`
    - Limiter : `GLOBAL`
    - Limit : `50`
    - Resets after : `300.0`

- permission_overwrite_delete
    - Endpoint : `/channels/{channel_id}/permissions/{overwrite_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `10`
    - Resets after : `15.0`

- permission_overwrite_create
    - Endpoint : `/channels/{channel_id}/permissions/{overwrite_id}`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `10`
    - Resets after : `15.0`

- channel_pin_get_all
    - Endpoint : `/channels/{channel_id}/pins`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `5`

- channel_pin_ack
    - Endpoint : `/channels/{channel_id}/pins/ack`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- message_unpin
    - Endpoint : `/channels/{channel_id}/pins/{message_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `4`

- message_pin
    - Endpoint : `/channels/{channel_id}/pins/{message_id}`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `4`

- channel_group_user_get_all
    - Endpoint : `/channels/{channel_id}/recipients/`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- channel_group_user_delete
    - Endpoint : `/channels/{channel_id}/recipients/{user_id}`
    - Method : `DELETE`
    - Required auth : `user`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- channel_group_user_add
    - Endpoint : `/channels/{channel_id}/recipients/{user_id}`
    - Method : `PUT`
    - Required auth : `user`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- thread_leave
    - Endpoint : `/channels/{channel_id}/thread-members/@me`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- thread_user_get_all
    - Endpoint : `/channels/{channel_id}/thread-members`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `10`
    - Resets after : `10.0`

- thread_join
    - Endpoint : `/channels/{channel_id}/thread-members/@me`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- thread_user_delete
    - Endpoint : `/channels/{channel_id}/thread-members/{user_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- thread_user_get
    - Endpoint : `/channels/{channel_id}/thread-members/{user_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- thread_user_add
    - Endpoint : `/channels/{channel_id}/thread-members/{user_id}`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- thread_self_settings_edit
    - Endpoint : `/channels/{channel_id}/thread-members/@me/settings`
    - Method : `PATCH`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `4`
    - Resets after : `60.0`

- thread_create
    - Endpoint : `/channels/{channel_id}/threads`
    - Method : `POST`
    - Required auth : `UN`
    - Limiter : `GLOBAL`
    - Limit : `50`
    - Resets after : `300.0`

- channel_thread_get_chunk_active
    - Endpoint : `/channels/{channel_id}/threads/active`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- channel_thread_get_chunk_archived_private
    - Endpoint : `/channels/{channel_id}/threads/archived/private`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- channel_thread_get_chunk_archived_public
    - Endpoint : `/channels/{channel_id}/threads/archived/public`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- typing
    - Endpoint : `/channels/{channel_id}/typing`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `5`
    - Resets after : `5.0`

- channel_thread_get_chunk_self_archived
    - Endpoint : `/channels/{channel_id}/users/@me/threads/archived/private`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- webhook_get_all_channel
    - Endpoint : `/channels/{channel_id}/webhooks`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- webhook_create
    - Endpoint : `/channels/{channel_id}/webhooks`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `channel_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- discovery_stage_get_all
    - Endpoint : `/discovery`
    - Method : `GET`
    - Required auth : `UN`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`
    - Notes : Untested. `DiscordException Not Found (404): 404: Not Found`

- discovery_category_get_all
    - Endpoint : `/discovery/categories`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `10`
    - Resets after : `120.0`

- discovery_validate_term
    - Endpoint : `/discovery/valid-term`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `10`
    - Resets after : `10.0`

- client_gateway_hooman
    - Endpoint : `/gateway`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `UN`
    - Resets after : `UN`
    - Notes : Untested.

- client_gateway_bot
    - Endpoint : `/gateway/bot`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `2`
    - Resets after : `5`

- discovery_guild_get_all
    - Endpoint : `/guild-discovery`
    - Method : `GET`
    - Required auth : `UN`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`
    - Notes : Untested. `DiscordException Not Found (404): 404: Not Found`

- scheduled_event_delete
    - Endpoint : `/guild-events/{scheduled_event_id}`
    - Method : `DELETE`
    - Required auth : `UN`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`

- scheduled_event_get
    - Endpoint : `/guild-events/{scheduled_event_id}`
    - Method : `GET`
    - Required auth : `UN`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`

- scheduled_event_edit
    - Endpoint : `/guild-events/{scheduled_event_id}`
    - Method : `PATCH`
    - Required auth : `UN`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`

- guild_create
    - Endpoint : `/guilds`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- guild_delete
    - Endpoint : `/guilds/{guild_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- guild_get
    - Endpoint : `/guilds/{guild_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_edit
    - Endpoint : `/guilds/{guild_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_ack
    - Endpoint : `/guilds/{guild_id}/ack`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `UN`
    - Resets after : `UN`
    - Notes : Untested.

- audit_log_get_chunk
    - Endpoint : `/guilds/{guild_id}/audit-logs`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_ban_get_all
    - Endpoint : `/guilds/{guild_id}/bans`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_ban_delete
    - Endpoint : `/guilds/{guild_id}/bans/{user_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_ban_get
    - Endpoint : `/guilds/{guild_id}/bans/{user_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_ban_add
    - Endpoint : `/guilds/{guild_id}/bans/{user_id}`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_channel_get_all
    - Endpoint : `/guilds/{guild_id}/channels`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- channel_move
    - Endpoint : `/guilds/{guild_id}/channels`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- channel_create
    - Endpoint : `/guilds/{guild_id}/channels`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_discovery_delete_subcategory
    - Endpoint : `/guilds/{guild_id}/discovery-categories/{category_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_discovery_add_subcategory
    - Endpoint : `/guilds/{guild_id}/discovery-categories/{category_id}`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_discovery_get
    - Endpoint : `/guilds/{guild_id}/discovery-metadata`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_discovery_edit
    - Endpoint : `/guilds/{guild_id}/discovery-metadata`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_embed_get
    - Endpoint : `/guilds/{guild_id}/embed`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Deprecated. Works on v6, v7.

- guild_embed_edit
    - Endpoint : `/guilds/{guild_id}/embed`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Deprecated. Works on v6, v7.

- guild.embed_url
    - Endpoint : `/guilds/{guild_id}/embed.png`
    - Method : `GET`
    - Required auth : `N/A`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`
    - Notes : Deprecated. Works on v6, v7.

- emoji_guild_get_all
    - Endpoint : `/guilds/{guild_id}/emojis`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- emoji_create
    - Endpoint : `/guilds/{guild_id}/emojis`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `50`
    - Resets after : `3600.0`
    - Notes : Creating an emoji with the same name of an existing one has a different rate limit.
              Please don't do that.

- emoji_delete
    - Endpoint : `/guilds/{guild_id}/emojis`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `2.0`

- emoji_get
    - Endpoint : `/guilds/{guild_id}/emojis/{emoji_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- emoji_edit
    - Endpoint : `/guilds/{guild_id}/emojis/{emoji_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `2.0`

- scheduled_event_get_all_guild
    - Endpoint : `/guilds/{guild_id}/events`
    - Method : `GET`
    - Required auth : `UN`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`

- scheduled_event_create
    - Endpoint : `/guilds/{guild_id}/events`
    - Method : `POST`
    - Required auth : `UN`
    - Limiter : `UN`
    - Limit : `UN`
    - Resets after : `UN`

- integration_get_all
    - Endpoint : `/guilds/{guild_id}/integrations`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- integration_create
    - Endpoint : `/guilds/{guild_id}/integrations`
    - Method : `POST`
    - Required auth : `UN`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- integration_delete
    - Endpoint : `/guilds/{guild_id}/integrations/{integration_id}`
    - Method : `DELETE`
    - Required auth : `UN`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- integration_edit
    - Endpoint : `/guilds/{guild_id}/integrations/{integration_id}`
    - Method : `PATCH`
    - Required auth : `UN`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- integration_edit
    - Endpoint : `/guilds/{guild_id}/integrations/{integration_id}/sync`
    - Method : `POST`
    - Required auth : `UN`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- invite_get_all_guild
    - Endpoint : `/guilds/{guild_id}/invites`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- verification_screen_get
    - Endpoint : `/guilds/{guild_id}/member-verification`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- verification_screen_edit
    - Endpoint : `/guilds/{guild_id}/member-verification`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_user_get_chunk
    - Endpoint : `/guilds/{guild_id}/members`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- client_guild_profile_edit
    - Endpoint : `/guilds/{guild_id}/members/@me`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `1.0`

- client_guild_profile_nick_edit
    - Endpoint : `/guilds/{guild_id}/members/@me/nick`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `1.0`

- guild_user_delete
    - Endpoint : `/guilds/{guild_id}/members/{user_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `5`
    - Resets after : `1.0`

- guild_user_get
    - Endpoint : `/guilds/{guild_id}/members/{user_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `5`
    - Resets after : `1.0`

- user_guild_profile_edit, user_move
    - Endpoint : `/guilds/{guild_id}/members/{user_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- guild_user_add
    - Endpoint : `/guilds/{guild_id}/members/{user_id}`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- guild_user_search
    - Endpoint : `/guilds/{guild_id}/members/search`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- user_role_delete
    - Endpoint : `/guilds/{guild_id}/members/{user_id}/roles/{role_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- user_role_add
    - Endpoint : `/guilds/{guild_id}/members/{user_id}/roles/{role_id}`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `10`
    - Resets after : `10.0`

- guild_preview_get
    - Endpoint : `/guilds/{guild_id}/preview`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- guild_prune_estimate
    - Endpoint : `/guilds/{guild_id}/prune`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_prune
    - Endpoint : `/guilds/{guild_id}/prune`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_voice_region_get_all
    - Endpoint : `/guilds/{guild_id}/regions`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_role_get_all
    - Endpoint : `/guilds/{guild_id}/roles`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- role_move
    - Endpoint : `/guilds/{guild_id}/roles`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- role_create
    - Endpoint : `/guilds/{guild_id}/roles`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `250`
    - Resets after : `172800.0`

- role_delete
    - Endpoint : `/guilds/{guild_id}/roles/{role_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- role_edit
    - Endpoint : `/guilds/{guild_id}/roles/{role_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `1000`
    - Resets after : `86400.0`

- sticker_guild_get_all
    - Endpoint : `/guilds/{guild_id}/stickers`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- sticker_guild_create
    - Endpoint : `/guilds/{guild_id}/stickers`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `50`
    - Resets after : `3600.0`

- sticker_guild_edit
    - Endpoint : `/guilds/{guild_id}/stickers/{sticker_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `2.0`

- sticker_guild_delete
    - Endpoint : `/guilds/{guild_id}/stickers/{sticker_id}`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `2.0`

- sticker_guild_get
    - Endpoint : `/guilds/{guild_id}/stickers/{sticker_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- guild_thread_get_all_active
    - Endpoint : `/guilds/{guild_id}/threads/active`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- vanity_invite_get
    - Endpoint : `/guilds/{guild_id}/vanity-url`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- vanity_invite_edit
    - Endpoint : `/guilds/{guild_id}/vanity-url`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `10`
    - Resets after : `60.0`

- voice_state_client_edit
    - Endpoint : `/guilds/{guild_id}/voice-states/@me`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : GLOBAL
    - Limit : 15
    - Resets after : `5.0`

- voice_state_user_edit
    - Endpoint : `/guilds/{guild_id}/voice-states/@me`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : GLOBAL
    - Limit : `5`
    - Resets after : `5.0`

- welcome_screen_get
    - Endpoint : `/guilds/{guild_id}/welcome-screen`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- welcome_screen_edit
    - Endpoint : `/guilds/{guild_id}/welcome-screen`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `100`
    - Resets after : `86400.0`

- webhook_get_all_guild
    - Endpoint : `/guilds/{guild_id}/webhooks`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `guild_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_widget_get
    - Endpoint : `/guilds/{guild_id}/widget.json`
    - Method : `GET`
    - Required auth : `N/A`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- guild.widget_url
    - Endpoint : `/guilds/{guild_id}/widget.png`
    - Method : `GET`
    - Required auth : `N/A`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- hypesquad_house_leave
    - Endpoint : `/hypesquad/online`
    - Method : `DELETE`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `UN`
    - Resets after : `UN`
    - Notes : Untested.

- hypesquad_house_change
    - Endpoint : `/hypesquad/online`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `UN`
    - Resets after : `UN`
    - Notes : Untested.

- interaction_response_message_create
    - Endpoint : `/interactions/{interaction_id}/{interaction_token}/callback`
    - Method : `POST`
    - Required auth : `N/A`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- invite_delete
    - Endpoint : `/invites/{invite_code}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- invite_get
    - Endpoint : `/invites/{invite_code}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `250`
    - Resets after : `6.0`

- client_application_get
    - Endpoint : `/oauth2/applications/@me`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- bulk_ack
    - Endpoint : `/read-states/ack-bulk`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `UN`
    - Resets after : `UN`
    - Notes : Untested.

- stage_get_all
    - Endpoint : `/stage-instances`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`
    - Notes : `DiscordException Forbidden (403), code=20001: Bots cannot use this endpoint`

- stage_create
    - Endpoint : `/stage-instances`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `100`
    - Resets after : `86400.0`

- stage_get
    - Endpoint : `/stage-instances/{channel_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- stage_delete
    - Endpoint : `/stage-instances/{channel_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `100`
    - Resets after : `86400.0`

- stage_edit
    - Endpoint : `/stage-instances/{channel_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `60.0`

- sticker_pack_get_all
    - Endpoint : `/sticker-packs`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- sticker_get
    - Endpoint : `/stickers/{sticker_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `2.0`

- eula_get
    - Endpoint : `/store/eulas/{eula_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- user_info_get
    - Endpoint : `/users/@me`
    - Method : `GET`
    - Required auth : `bearer`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- client_user_get
    - Endpoint : `/users/@me`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- client_edit
    - Endpoint : `/users/@me`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `2`
    - Resets after : `600.0`

- user_achievement_get_all
    - Endpoint : `/users/@me/applications/{application_id}/achievements`
    - Method : `GET`
    - Required auth : `bearer`
    - Limiter : `GLOBAL`
    - Limit : `2`
    - Resets after : `5.0`
    - Notes : Untested.

- user_achievement_update
    - Endpoint : `/users/{user_id}/applications/{application_id}/achievements/{achievement_id}`
    - Method : `PUT`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `5`
    - Resets after : `5.0`

- channel_private_get_all
    - Endpoint : `/users/@me/channels`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- channel_private_create
    - Endpoint : `/users/@me/channels`
    - Method : `POST`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- client_connection_get_all
    - Endpoint : `/users/@me/connections`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- user_connection_get_all
    - Endpoint : `/users/@me/connections`
    - Method : `GET`
    - Required auth : `bearer`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- guild_get_all
    - Endpoint : `/users/@me/guilds`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `1.0`

- user_guild_get_all
    - Endpoint : `/users/@me/guilds`
    - Method : `GET`
    - Required auth : `bearer`
    - Limiter : `GLOBAL`
    - Limit : `1`
    - Resets after : `1.0`

- guild_leave
    - Endpoint : `/users/@me/guilds/{guild_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- relationship_friend_request
    - Endpoint : `/users/@me/relationships`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- relationship_delete
    - Endpoint : `/users/@me/relationships/{user_id}`
    - Method : `DELETE`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- relationship_create
    - Endpoint : `/users/@me/settings`
    - Method : `PUT`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- client_settings_get
    - Endpoint : `/users/@me/settings`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- client_settings_edit
    - Endpoint : `/users/@me/settings`
    - Method : `PATCH`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- user_get
    - Endpoint : `/users/{user_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `30`
    - Resets after : `30.0`

- channel_group_create
    - Endpoint : `/users/{user_id}/channels`
    - Method : `POST`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- user_get_profile
    - Endpoint : `users/{user_id}/profile`
    - Method : `GET`
    - Required auth : `user`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`
    - Notes : Untested.

- voice_region_get_all
    - Endpoint : `/voice/regions`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `GLOBAL`
    - Limit : `OPT`
    - Resets after : `OPT`

- interaction_followup_message_create
    - Endpoint : `webhooks/{application_id}/{interaction_token}`
    - Method : `POST`
    - Required auth : `N/A`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- interaction_response_message_delete
    - Endpoint : `webhooks/{application_id}/{interaction_token}/messages/@original`
    - Method : `DELETE`
    - Required auth : `N/A`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- interaction_response_message_get
    - Endpoint : `webhooks/{application_id}/{interaction_token}/messages/@original`
    - Method : `GET`
    - Required auth : `N/A`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- interaction_response_message_edit
    - Endpoint : `webhooks/{application_id}/{interaction_token}/messages/@original`
    - Method : `PATCH`
    - Required auth : `N/A`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- interaction_followup_message_delete
    - Endpoint : `webhooks/{application_id}/{interaction_token}/messages/{message_id}`
    - Method : `DELETE`
    - Required auth : `N/A`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- interaction_followup_message_get
    - Endpoint : `webhooks/{application_id}/{interaction_token}/messages/{message_id}`
    - Method : `GET`
    - Required auth : `N/A`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- interaction_followup_message_edit
    - Endpoint : `webhooks/{application_id}/{interaction_token}/messages/{message_id}`
    - Method : `PATCH`
    - Required auth : `N/A`
    - Limiter : `interaction_id`
    - Limit : `5`
    - Resets after : `2.0`

- webhook_delete
    - Endpoint : `/webhooks/{webhook_id}`
    - Method : `DELETE`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- webhook_get
    - Endpoint : `/webhooks/{webhook_id}`
    - Method : `GET`
    - Required auth : `bot`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- webhook_edit
    - Endpoint : `/webhooks/{webhook_id}`
    - Method : `PATCH`
    - Required auth : `bot`
    - Limiter : `webhook_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- webhook_delete_token
    - Endpoint : `/webhooks/{webhook_id}/{webhook_token}`
    - Method : `DELETE`
    - Required auth : `N/A`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- webhook_get_token
    - Endpoint : `/webhooks/{webhook_id}/{webhook_token}`
    - Method : `GET`
    - Required auth : `N/A`
    - Limiter : `UNLIMITED`
    - Limit : `N/A`
    - Resets after : `N/A`

- webhook_edit_token
    - Endpoint : `/webhooks/{webhook_id}/{webhook_token}`
    - Method : `PATCH`
    - Required auth : `N/A`
    - Limiter : `webhook_id`
    - Limit : `OPT`
    - Resets after : `OPT`

- webhook_message_create
    - Endpoint : `/webhooks/{webhook_id}/{webhook_token}`
    - Method : `POST`
    - Required auth : `N/A`
    - Limiter : `webhook_id`
    - Limit : `5`
    - Resets after : `2.0`

- webhook_message_delete
    - Endpoint : `/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}`
    - Method : `DELETE`
    - Required auth : `N/A`
    - Limiter : `webhook_id`
    - Limit : `5`
    - Resets after : `2.0`

- webhook_message_get
    - Endpoint : `/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}`
    - Method : `GET`
    - Required auth : `N/A`
    - Limiter : `webhook_id`
    - Limit : `5`
    - Resets after : `2.0`

- webhook_message_edit
    - Endpoint : `/webhooks/{webhook_id}/{webhook_token}/messages/{message_id}`
    - Method : `PATCH`
    - Required auth : `N/A`
    - Limiter : `webhook_id`
    - Limit : `5`
    - Resets after : `2.0`
"""
GROUP_REACTION_MODIFY = RateLimitGroup(LIMITER_CHANNEL)
GROUP_PIN_MODIFY = RateLimitGroup(LIMITER_CHANNEL)
GROUP_USER_MODIFY = RateLimitGroup(LIMITER_GUILD) # both has the same endpoint
GROUP_USER_ROLE_MODIFY = RateLimitGroup(LIMITER_GUILD)
GROUP_WEBHOOK_EXECUTE = RateLimitGroup(LIMITER_WEBHOOK)
GROUP_INTERACTION_EXECUTE = RateLimitGroup(LIMITER_INTERACTION)
GROUP_APPLICATION_COMMAND_CREATE = RateLimitGroup()
GROUP_APPLICATION_COMMAND_DELETE = RateLimitGroup()
GROUP_APPLICATION_COMMAND_EDIT = RateLimitGroup()
GROUP_PERMISSION_OVERWRITE_MODIFY = RateLimitGroup(LIMITER_CHANNEL)
GROUP_THREAD_CREATE = RateLimitGroup()
GROUP_THREAD_ACTION = RateLimitGroup()

oauth2_token = RateLimitGroup(optimistic=True)
application_get = RateLimitGroup(optimistic=True) # untested
achievement_get_all = RateLimitGroup()
achievement_create = RateLimitGroup()
achievement_delete = RateLimitGroup()
achievement_get = RateLimitGroup()
achievement_edit = RateLimitGroup()
application_command_global_get_all = RateLimitGroup.unlimited()
application_command_global_delete = RateLimitGroup()
application_command_global_create = RateLimitGroup()
application_command_global_update_multiple = RateLimitGroup()
application_command_global_get = RateLimitGroup.unlimited()
application_command_global_edit = RateLimitGroup()
application_command_guild_get_all = RateLimitGroup.unlimited()
application_command_guild_update_multiple = RateLimitGroup(LIMITER_GUILD)
application_command_permission_get_all_guild = RateLimitGroup.unlimited()
application_command_guild_delete = RateLimitGroup(LIMITER_GUILD)
application_command_guild_create = RateLimitGroup(LIMITER_GUILD)
application_command_guild_get = RateLimitGroup.unlimited()
application_command_guild_edit = RateLimitGroup(LIMITER_GUILD)
application_command_permission_get = RateLimitGroup.unlimited()
application_command_permission_edit = RateLimitGroup()
application_get_all_detectable = RateLimitGroup(optimistic=True)
client_logout = RateLimitGroup() # untested
channel_delete = RateLimitGroup.unlimited()
channel_group_leave = RateLimitGroup.unlimited() # untested; same as channel_delete?
channel_edit = RateLimitGroup(LIMITER_CHANNEL)
channel_group_edit = RateLimitGroup(LIMITER_CHANNEL, optimistic=True) # untested; same as channel_edit?
channel_directory_counts = RateLimitGroup(LIMITER_CHANNEL, optimistic=True) # untested; same as channel_edit?
channel_directory_get_all = RateLimitGroup(LIMITER_CHANNEL, optimistic=True) # untested; same as channel_edit?
channel_directory_search = RateLimitGroup(LIMITER_CHANNEL, optimistic=True) # untested; same as channel_edit?
channel_follow = RateLimitGroup(LIMITER_CHANNEL, optimistic=True)
greet = RateLimitGroup(optimistic=True) # untested
invite_get_all_channel = RateLimitGroup(LIMITER_CHANNEL, optimistic=True)
invite_create = RateLimitGroup()
message_get_chunk = RateLimitGroup(LIMITER_CHANNEL)
message_create = RateLimitGroup(LIMITER_CHANNEL)
message_delete_multiple = RateLimitGroup(LIMITER_CHANNEL)
message_delete = RateLimitGroup(LIMITER_CHANNEL)
message_delete_b2wo = RateLimitGroup(LIMITER_CHANNEL)
message_get = RateLimitGroup(LIMITER_CHANNEL)
message_edit = RateLimitGroup(LIMITER_CHANNEL)
message_ack = RateLimitGroup(optimistic=True) # untested
message_crosspost = RateLimitGroup(LIMITER_CHANNEL)
message_interaction = RateLimitGroup(optimistic=True) # untested
reaction_clear = GROUP_REACTION_MODIFY
reaction_delete_emoji = GROUP_REACTION_MODIFY
reaction_user_get_chunk = RateLimitGroup(LIMITER_CHANNEL, optimistic=True)
reaction_delete_own = GROUP_REACTION_MODIFY
reaction_add = GROUP_REACTION_MODIFY
reaction_delete = GROUP_REACTION_MODIFY
message_suppress_embeds = RateLimitGroup()
thread_create_from_message = GROUP_THREAD_CREATE
permission_overwrite_delete = GROUP_PERMISSION_OVERWRITE_MODIFY
permission_overwrite_create = GROUP_PERMISSION_OVERWRITE_MODIFY
channel_pin_get_all = RateLimitGroup()
channel_pin_ack = RateLimitGroup(optimistic=True) # untested
message_unpin = GROUP_PIN_MODIFY
message_pin = GROUP_PIN_MODIFY
channel_group_user_get_all = RateLimitGroup(LIMITER_CHANNEL, optimistic=True) # untested
channel_group_user_delete = RateLimitGroup(LIMITER_CHANNEL, optimistic=True) # untested
channel_group_user_add = RateLimitGroup(LIMITER_CHANNEL, optimistic=True) # untested
thread_user_get_all = RateLimitGroup()
thread_join = GROUP_THREAD_ACTION
thread_leave = GROUP_THREAD_ACTION
thread_user_add = GROUP_THREAD_ACTION
thread_user_delete = GROUP_THREAD_ACTION
thread_user_get = RateLimitGroup.unlimited()
thread_self_settings_edit = RateLimitGroup()
thread_create = GROUP_THREAD_CREATE
typing = RateLimitGroup(LIMITER_CHANNEL)
channel_thread_get_chunk_active = RateLimitGroup.unlimited()
channel_thread_get_chunk_archived_private = RateLimitGroup.unlimited()
channel_thread_get_chunk_archived_public = RateLimitGroup.unlimited()
channel_thread_get_chunk_self_archived = RateLimitGroup.unlimited()
webhook_get_all_channel = RateLimitGroup(LIMITER_CHANNEL, optimistic=True)
webhook_create = RateLimitGroup(LIMITER_CHANNEL, optimistic=True)
discovery_category_get_all = RateLimitGroup()
discovery_stage_get_all = RateLimitGroup(optimistic=True) # untested, not yet added
discovery_validate_term = RateLimitGroup()
client_gateway_hooman = RateLimitGroup()
client_gateway_bot = RateLimitGroup()
discovery_guild_get_all = RateLimitGroup(optimistic=True) # untested, not yet added
scheduled_event_delete = RateLimitGroup(optimistic=True) # untested
scheduled_event_get = RateLimitGroup(optimistic=True) # untested
scheduled_event_edit = RateLimitGroup(optimistic=True) # untested
guild_create = RateLimitGroup.unlimited()
guild_delete = RateLimitGroup.unlimited()
guild_get = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_edit = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_ack = RateLimitGroup() # untested
audit_log_get_chunk = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_ban_get_all = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_ban_delete = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_ban_get = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_ban_add = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_channel_get_all = RateLimitGroup(LIMITER_GUILD, optimistic=True)
channel_move = RateLimitGroup(LIMITER_GUILD, optimistic=True)
channel_create = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_discovery_delete_subcategory = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_discovery_add_subcategory = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_discovery_get = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_discovery_edit = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_embed_get = RateLimitGroup(LIMITER_GUILD, optimistic=True) # deprecated
guild_embed_edit = RateLimitGroup(LIMITER_GUILD, optimistic=True) # deprecated
emoji_guild_get_all = RateLimitGroup(LIMITER_GUILD, optimistic=True)
emoji_create = RateLimitGroup()
emoji_delete = RateLimitGroup()
emoji_get = RateLimitGroup(LIMITER_GUILD, optimistic=True)
emoji_edit = RateLimitGroup()
scheduled_event_get_all_guild = RateLimitGroup(optimistic=True) # untested
scheduled_event_create = RateLimitGroup(optimistic=True) # untested
integration_get_all = RateLimitGroup(LIMITER_GUILD, optimistic=True)
integration_create = RateLimitGroup(optimistic=True) # untested
integration_delete = RateLimitGroup(optimistic=True) # untested
integration_edit = RateLimitGroup(optimistic=True) # untested
integration_sync = RateLimitGroup(optimistic=True) # untested
invite_get_all_guild = RateLimitGroup(LIMITER_GUILD, optimistic=True)
verification_screen_get = RateLimitGroup(LIMITER_GUILD, optimistic=True)
verification_screen_edit = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_user_get_chunk = RateLimitGroup(LIMITER_GUILD)
client_guild_profile_edit = RateLimitGroup()
client_guild_profile_nick_edit = RateLimitGroup()
guild_user_delete = RateLimitGroup(LIMITER_GUILD)
guild_user_get = RateLimitGroup(LIMITER_GUILD)
user_guild_profile_edit = GROUP_USER_MODIFY
user_move = GROUP_USER_MODIFY
guild_user_add = RateLimitGroup(LIMITER_GUILD)
guild_user_search = RateLimitGroup(LIMITER_GUILD)
user_role_delete = GROUP_USER_ROLE_MODIFY
user_role_add = GROUP_USER_ROLE_MODIFY
guild_preview_get = RateLimitGroup()
guild_prune_estimate = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_prune = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_voice_region_get_all = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_role_get_all = RateLimitGroup(LIMITER_GUILD, optimistic=True)
role_move = RateLimitGroup(LIMITER_GUILD, optimistic=True)
role_create = RateLimitGroup(LIMITER_GUILD)
role_delete = RateLimitGroup(LIMITER_GUILD, optimistic=True)
role_edit = RateLimitGroup(LIMITER_GUILD)
sticker_guild_get_all = RateLimitGroup.unlimited()
sticker_guild_create = RateLimitGroup()
sticker_guild_edit = RateLimitGroup()
sticker_guild_delete = RateLimitGroup()
sticker_guild_get = RateLimitGroup.unlimited()
guild_thread_get_all_active = RateLimitGroup.unlimited()
vanity_invite_get = RateLimitGroup.unlimited()
vanity_invite_edit = RateLimitGroup()
voice_state_client_edit = RateLimitGroup()
voice_state_user_edit = RateLimitGroup.unlimited() # untested
welcome_screen_get = RateLimitGroup(LIMITER_GUILD, optimistic=True)
welcome_screen_edit = RateLimitGroup(LIMITER_GUILD)
webhook_get_all_guild = RateLimitGroup(LIMITER_GUILD, optimistic=True)
guild_widget_get = RateLimitGroup.unlimited()
hypesquad_house_leave = RateLimitGroup() # untested
hypesquad_house_change = RateLimitGroup() # untested
interaction_response_message_create = RateLimitGroup.unlimited()
invite_delete = RateLimitGroup.unlimited()
invite_get = RateLimitGroup()
client_application_get = RateLimitGroup(optimistic=True)
bulk_ack = RateLimitGroup(optimistic=True) # untested
stage_get_all = RateLimitGroup.unlimited()
stage_create = RateLimitGroup()
stage_get = RateLimitGroup()
stage_delete = RateLimitGroup()
stage_edit = RateLimitGroup()
sticker_pack_get_all = RateLimitGroup.unlimited()
sticker_get = RateLimitGroup()
eula_get = RateLimitGroup(optimistic=True)
user_info_get = RateLimitGroup(optimistic=True)
client_user_get = RateLimitGroup(optimistic=True)
client_edit = RateLimitGroup()
user_achievement_get_all = RateLimitGroup() # untested; has expected global rate limit
user_achievement_update = RateLimitGroup()
channel_private_get_all = RateLimitGroup(optimistic=True)
channel_private_create = RateLimitGroup.unlimited()
client_connection_get_all = RateLimitGroup(optimistic=True)
user_connection_get_all = RateLimitGroup(optimistic=True)
guild_get_all = RateLimitGroup()
user_guild_get_all = RateLimitGroup()
guild_leave = RateLimitGroup.unlimited()
relationship_friend_request = RateLimitGroup(optimistic=True) # untested
relationship_delete = RateLimitGroup(optimistic=True) # untested
relationship_create = RateLimitGroup(optimistic=True) # untested
client_settings_get = RateLimitGroup(optimistic=True) # untested
client_settings_edit = RateLimitGroup(optimistic=True) # untested
user_get = RateLimitGroup()
channel_group_create = RateLimitGroup(optimistic=True) # untested
user_get_profile = RateLimitGroup(optimistic=True) # untested
voice_region_get_all = RateLimitGroup(optimistic=True)
interaction_followup_message_create = GROUP_INTERACTION_EXECUTE
interaction_response_message_delete = GROUP_INTERACTION_EXECUTE
interaction_response_message_get = GROUP_INTERACTION_EXECUTE
interaction_response_message_edit = GROUP_INTERACTION_EXECUTE
interaction_followup_message_delete = GROUP_INTERACTION_EXECUTE
interaction_followup_message_get = GROUP_INTERACTION_EXECUTE
interaction_followup_message_edit = GROUP_INTERACTION_EXECUTE
webhook_delete = RateLimitGroup.unlimited()
webhook_get = RateLimitGroup.unlimited()
webhook_edit = RateLimitGroup(LIMITER_WEBHOOK, optimistic=True)
webhook_delete_token = RateLimitGroup.unlimited()
webhook_get_token = RateLimitGroup.unlimited()
webhook_edit_token = RateLimitGroup(LIMITER_WEBHOOK, optimistic=True)
webhook_message_create = GROUP_WEBHOOK_EXECUTE
webhook_message_edit = GROUP_WEBHOOK_EXECUTE
webhook_message_get = GROUP_WEBHOOK_EXECUTE
webhook_message_delete = GROUP_WEBHOOK_EXECUTE

# Alternative static versions
STATIC_MESSAGE_DELETE_SUB = StaticRateLimitGroup(5, 5.0, LIMITER_CHANNEL)
static_message_delete = (STATIC_MESSAGE_DELETE_SUB, StaticRateLimitGroup(3, 1.0, LIMITER_CHANNEL))
static_message_delete_b2wo = (STATIC_MESSAGE_DELETE_SUB, StaticRateLimitGroup(30, 120.0, LIMITER_CHANNEL))


# Status
status_incident_unresolved = RateLimitGroup.unlimited()
status_maintenance_active = RateLimitGroup.unlimited()
status_maintenance_upcoming = RateLimitGroup.unlimited()
